"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from ANNarchy.core import Global
from ANNarchy.core.PopulationView import PopulationView
from ANNarchy.core.Population import Population
from ANNarchy.core.Projection import Projection

from ANNarchy.extensions.bold import BoldMonitor
from ANNarchy.extensions.convolution import Transpose

from ANNarchy.intern.NetworkManager import NetworkManager
from ANNarchy.intern.ConfigManagement import get_global_config, _check_paradigm
from ANNarchy.intern.GlobalObjects import GlobalObjectManager
from ANNarchy.intern.Profiler import Profiler

from ANNarchy.generator.Template import PyxTemplate
from ANNarchy.generator.Population import OpenMPTemplates as omp_templates
from ANNarchy.generator.Population import CUDATemplates as cuda_templates
from ANNarchy.generator.Projection.SingleThread import *
from ANNarchy.generator.Projection.OpenMP import *
from ANNarchy.generator.Projection.CUDA import *
from ANNarchy.generator.Utils import tabify, determine_idx_type_for_projection, cpp_connector_available

class PyxGenerator(object):
    """
    Generate the python extension (*.pyx) file comprising of wrapper
    classes for the individual objects. Secondly the definition of accessible
    methods, e. g. simulate(int steps). Generally an extension consists of two
    parts: a struct definition (define accessible parts of the C++ object) and
    the wrapper object.

    In detail, there are extensions available for:

        * parsed populations
        * parsed projections
        * recorder objects

    Implementation Note (HD: 17.06.2015)

        This class could be implemented in general as a set of functions.
        Nevertheless we chose an object-oriented approach, as it is easier to
        use from Generator object. As the submethods has no access to data
        stored in PyxGenerator (there is no real data at all) all private
        functions are marked with @staticmethod to make this clear.

    TODO:

        * handling of specific populations, projections is currently done over
        generator struct in the population/projection object. This should be
        changed as in the C++ code generation.
    """
    def __init__(self, annarchy_dir, populations, projections, net_id):
        """
        Store a list of population und projection objects for later processing.
        """
        self._annarchy_dir = annarchy_dir
        self._populations = populations
        self._projections = projections
        self._net_id = net_id

    def generate(self):
        """
        Perform code generation.
        """
        # Custom user-defined functions (add_function())
        custom_functions_export, functions_wrapper = self._custom_functions()

        # Custom user-defined constants (add_constant())
        custom_constants_export, constants_wrapper = self._custom_constants()

        # struct declaration for each population
        pop_struct = ""
        pop_ptr = ""
        for pop in self._populations:
            # Header export
            pop_struct += self._pop_struct(pop)
            # Population instance
            pop_ptr += """
    PopStruct%(id)s pop%(id)s"""% {
    'id': pop.id,
}

        # struct declaration for each projection
        proj_struct = ""
        proj_ptr = ""
        for proj in self._projections:
            # Header export
            proj_struct += self._proj_struct(proj)

            # Projection instance
            proj_ptr += """
    ProjStruct%(id_proj)s proj%(id_proj)s"""% {
    'id_proj': proj.id,
}

        # struct declaration for each monitor
        monitor_struct = "" #self._pyx_struct_monitor()
        for pop in self._populations:
            monitor_struct += self._pop_monitor_struct(pop)
        for proj in self._projections:
            monitor_struct += self._proj_monitor_struct(proj)
        for mon in NetworkManager().get_monitors(net_id=self._net_id):
            if isinstance(mon, BoldMonitor):
                mon_dict = {
                    'pop_id': mon.object.id,
                    'pop_name': mon.object.name,
                    'mon_id': mon.id,
                    'float_prec': get_global_config('precision')
                }
                monitor_struct += mon._specific_template['pyx_struct'] % mon_dict

        # Cython wrappers for the populations
        pop_class = ""
        for pop in self._populations:
            pop_class += self._pop_wrapper(pop)

        # Cython wrappers for the projections
        proj_class = ""
        for proj in self._projections:
            proj_class += self._proj_wrapper(proj)

        # Cython wrappers for the monitors
        monitor_class = ""
        for pop in self._populations:
            monitor_class += self._pop_monitor_wrapper(pop)
        for proj in self._projections:
            monitor_class += self._proj_monitor_wrapper(proj)
        for mon in NetworkManager().get_monitors(net_id=self._net_id):
            if isinstance(mon, BoldMonitor):
                mon_dict = {
                    'pop_id': mon.object.id,
                    'pop_name': mon.object.name,
                    'mon_id': mon.id,
                    'float_prec': get_global_config('precision')
                }
                monitor_class += mon._specific_template['pyx_wrapper'] % mon_dict

        if Profiler().enabled:
            prof_class = PyxTemplate.pyx_profiler_template
        else:
            prof_class = ""

        from .Template.PyxTemplate import pyx_template
        return pyx_template % {
            'custom_functions_export': custom_functions_export,
            'custom_constants_export': custom_constants_export,
            'prof_class': prof_class,
            'pop_struct': pop_struct,
            'pop_ptr': pop_ptr,
            'proj_struct': proj_struct,
            'proj_ptr': proj_ptr,
            'pop_class' : pop_class,
            'proj_class': proj_class,
            'monitor_struct': monitor_struct,
            'monitor_wrapper': monitor_class,
            'functions_wrapper': functions_wrapper,
            'constants_wrapper': constants_wrapper,
            'float_prec': get_global_config('precision'),
            'device_specific_export': PyxTemplate.pyx_device_specific[get_global_config('paradigm')]['export'],
            'device_specific_wrapper': PyxTemplate.pyx_device_specific[get_global_config('paradigm')]['wrapper'],
        }

    @staticmethod
    def _get_proj_template(proj):
        """
        Choose the correct template collection dependent on target platform and storage formate.

        For example the export of delay related variables.
        """
        if get_global_config('paradigm') == "openmp":
            if proj._storage_format == "lil":
                if get_global_config('num_threads') == 1:
                    return LIL_SingleThread.conn_templates
                else:
                    if proj._no_split_matrix:
                        return LIL_OpenMP.conn_templates
                    else:
                        return LIL_Sliced_OpenMP.conn_templates

            elif proj._storage_format == "coo":
                if get_global_config('num_threads') == 1:
                    return COO_SingleThread.conn_templates
                else:
                    return COO_OpenMP.conn_templates

            elif proj._storage_format == "dia":
                if get_global_config('num_threads') == 1:
                    return DIA_SingleThread.conn_templates
                else:
                    return DIA_OpenMP.conn_templates

            elif proj._storage_format == "bsr":
                if get_global_config('num_threads') == 1:
                    return BSR_SingleThread.conn_templates
                else:
                    return BSR_OpenMP.conn_templates

            elif proj._storage_format == "csr":
                if get_global_config('num_threads') == 1:
                    return CSR_SingleThread.conn_templates
                else:
                    return CSR_OpenMP.conn_templates

            elif proj._storage_format == "ellr":
                if get_global_config('num_threads') == 1:
                    return ELLR_SingleThread.conn_templates
                else:
                    return ELLR_OpenMP.conn_templates

            elif proj._storage_format == "sell":
                if get_global_config('num_threads') == 1:
                    return SELL_SingleThread.conn_templates
                else:
                    return SELL_OpenMP.conn_templates

            elif proj._storage_format == "ell":
                if get_global_config('num_threads') == 1:
                    return ELL_SingleThread.conn_templates
                else:
                    return ELL_OpenMP.conn_templates

            elif proj._storage_format == "hyb":
                if get_global_config('num_threads') == 1:
                    return HYB_SingleThread.conn_templates
                else:
                    raise NotImplementedError

            elif proj._storage_format == "dense":
                if get_global_config('num_threads') == 1:
                    return Dense_SingleThread.conn_templates
                else:
                    return Dense_OpenMP.conn_templates

            else:
                raise Global.InvalidConfiguration("    No python extension definition available for format = "+str(proj._storage_format)+" on CPUs")

        elif get_global_config('paradigm') == "cuda":
            if proj._storage_format == "bsr":
                return BSR_CUDA.conn_templates
            elif proj._storage_format == "csr":
                return CSR_CUDA.conn_templates
            elif proj._storage_format == "csr_scalar":
                return CSR_SCALAR_CUDA.conn_templates
            elif proj._storage_format == "csr_vector":
                return CSR_VECTOR_CUDA.conn_templates
            elif proj._storage_format == "coo":
                return COO_CUDA.conn_templates
            elif proj._storage_format == "sell":
                return SELL_CUDA.conn_templates
            elif proj._storage_format == "ellr":
                return ELLR_CUDA.conn_templates
            elif proj._storage_format == "ell":
                return ELL_CUDA.conn_templates
            elif proj._storage_format == "hyb":
                return HYB_CUDA.conn_templates
            elif proj._storage_format == "dense":
                return Dense_CUDA.conn_templates
            else:
                raise Global.InvalidConfiguration("    No python extension definition available for format = "+str(proj._storage_format)+" on GPUs")

        else:
            raise NotImplementedError

#######################################################################
############## Functions #############################################
#######################################################################
    @staticmethod
    def _custom_functions(obj=None):
        """
        Generate the Python extension code (export and the wrapper code) dependent
        on the type provided in *obj*.
        """
        desc_list = []

        # Check if there are functions where code must be generated
        if obj is None:
            if GlobalObjectManager().number_functions() == 0:
                return "", ""

            from ANNarchy.parser.Extraction import extract_functions
            for _, func in GlobalObjectManager().get_functions():
                desc_list.append(extract_functions(func, local_global=True)[0])
            wrapper_prefix = ""
            export_prefix = "func_"

        elif isinstance(obj, Population):
            if (len(obj.neuron_type.description['functions']) == 0):
                return "", ""
            desc_list = obj.neuron_type.description['functions']
            wrapper_prefix = "pop%(id)s." % {'id': obj.id}
            export_prefix = ""

        elif isinstance(obj, Projection):
            if len(obj.synapse_type.description['functions']) == 0:
                return "", ""
            desc_list = obj.synapse_type.description['functions']
            wrapper_prefix = "proj%(id)s." % {'id': obj.id}
            export_prefix = ""

        # Generate the code
        export = ""
        wrapper = ""
        for desc in desc_list:
            # Export
            export += desc['return_type'] + " " + desc['name'] + '('
            for idx, arg in enumerate(desc['arg_types']):
                export += arg
                if idx < len(desc['arg_types']) - 1:
                    export += ', '
            export += ')' + '\n'

            # Wrapper
            arguments=""
            wrapper += "cpdef np.ndarray " + export_prefix + desc['name'] + '('
            if obj is not None:
                wrapper += "self, "
            for idx, arg in enumerate(desc['args']):
                # Function call
                wrapper += arg
                if idx < len(desc['args']) - 1:
                    wrapper += ', '
                # Element access
                arguments += arg + "[i]"
                if idx < len(desc['args']) - 1:
                    arguments += ', '
            wrapper += '):'
            wrapper += """
    return np.array([%(wrapper_prefix)s%(funcname)s(%(args)s) for i in range(len(%(first_arg)s))])
""" % {'wrapper_prefix': wrapper_prefix, 'funcname': desc['name'], 'first_arg' : desc['args'][0], 'args': arguments}

        # Tabs depend on type
        if obj is None:
            export = tabify(export, 1)
        else:
            export = tabify(export, 2)
            wrapper = tabify(wrapper, 1)

        return export, wrapper

#######################################################################
############## Constants  #############################################
#######################################################################
    def _custom_constants(self):
        if GlobalObjectManager().number_constants() == 0:
            return "", ""

        export = ""
        wrapper = ""
        for obj in GlobalObjectManager().get_constants():
            export += """
    void set_%(name)s(%(float_prec)s)""" % {'name': obj.name, 'float_prec': get_global_config('precision')}
            wrapper += """
def _set_%(name)s(%(float_prec)s value):
    set_%(name)s(value)""" % {'name': obj.name, 'float_prec': get_global_config('precision')}

        return export, wrapper


#######################################################################
############## Population #############################################
#######################################################################
    @staticmethod
    def _pop_struct(pop):
        """
        Generate population struct definition, mimics the c++ class.
        """

        # Spiking neurons have additional data
        export_refractory = ""
        if pop.neuron_type.type == 'spike':
            if pop.neuron_type.refractory or pop.refractory:
                if get_global_config('paradigm') == "openmp":
                    export_refractory = omp_templates.spike_specific["refractory"]["pyx_export"]
                elif get_global_config('paradigm') == "cuda":
                    export_refractory = cuda_templates.spike_specific["refractory"]["pyx_export"]
                else:
                    raise NotImplementedError

        export_parameters_variables = ""
        datatypes = PyxGenerator._get_datatypes(pop)
        # Local parameters and variables
        for ctype in datatypes["local"]:
            export_parameters_variables += PyxTemplate.pyx_default_pop_attribute_export["local"] % {
                'ctype': ctype,
                'ctype_name': ctype.replace(" ", "_")
            }

        # Global parameters and variables
        for ctype in datatypes["global"]:
            export_parameters_variables += PyxTemplate.pyx_default_pop_attribute_export["global"] % {
                'ctype': ctype,
                'ctype_name': ctype.replace(" ", "_")
            }

        if 'export_parameters_variables' in pop._specific_template.keys():
            export_parameters_variables = pop._specific_template['export_parameters_variables']

        # Local functions
        export_functions, _ = PyxGenerator._custom_functions(pop)

        # Mean firing rate
        export_mean_fr = ""
        if pop.neuron_type.type == 'spike':
            export_mean_fr = """
        # Compute firing rate
        void compute_firing_rate(%(float_prec)s window)""" %{'float_prec': get_global_config('precision')}

        # Additional exports
        export_additional = ""
        if 'export_additional' in pop._specific_template.keys():
            export_additional = pop._specific_template['export_additional']

        # Finalize the code
        return PyxTemplate.pop_pyx_struct % {
            'id': pop.id, 'name': pop.name,
            'export_refractory': export_refractory,
            'export_parameters_variables': export_parameters_variables,
            'export_functions': export_functions,
            'export_mean_fr': export_mean_fr,
            'export_additional': export_additional,
        }


    @staticmethod
    def _pop_wrapper(pop):
        """
        Generate population wrapper definition.
        """
        wrapper_args = "size, max_delay"
        wrapper_init = """
        pop%(id)s.set_size(size)
        pop%(id)s.set_max_delay(max_delay)""" % {'id': pop.id}
        wrapper_access_parameters_variables = ""
        wrapper_access_refractory = ""
        wrapper_access_additional = ""


        # Spiking neurons have aditional data
        if pop.neuron_type.type == 'spike':
            if pop.neuron_type.refractory or pop.refractory:
                if get_global_config('paradigm') == "openmp":
                    wrapper_access_refractory += omp_templates.spike_specific['refractory']['pyx_wrapper'] % {'id': pop.id}
                elif get_global_config('paradigm') == "cuda":
                    wrapper_access_refractory += cuda_templates.spike_specific['refractory']['pyx_wrapper'] % {'id': pop.id}
                else:
                    raise NotImplementedError

        # Attributes
        wrapper_access_parameters_variables = PyxGenerator._pop_generate_default_wrapper(pop)

        # Local functions
        _, wrapper_access_functions = PyxGenerator._custom_functions(pop)

        # Mean firing rate
        wrapper_access_mean_fr = ""
        if pop.neuron_type.type == 'spike':
            wrapper_access_mean_fr = """
    # Compute firing rate
    cpdef compute_firing_rate(self, double window):
        pop%(id)s.compute_firing_rate(window)"""% {'id': pop.id}

        # Specific populations can overwrite
        if 'wrapper_args' in pop._specific_template.keys():
            wrapper_args = pop._specific_template['wrapper_args']
        if 'wrapper_init' in pop._specific_template.keys():
            wrapper_init = pop._specific_template['wrapper_init']
        if 'wrapper_access_refractory' in pop._specific_template.keys():
            wrapper_access_refractory = pop._specific_template['wrapper_access_refractory']
        if 'wrapper_access_parameters_variables' in pop._specific_template.keys():
            wrapper_access_parameters_variables = pop._specific_template['wrapper_access_parameters_variables']
        if 'wrapper_access_additional' in pop._specific_template.keys():
            wrapper_access_additional = pop._specific_template['wrapper_access_additional']

        # Finalize the code
        return PyxTemplate.pop_pyx_wrapper % {
            'id': pop.id, 'name': pop.name,
            'wrapper_args' : wrapper_args,
            'wrapper_init' : wrapper_init,
            'wrapper_access_parameters_variables' : wrapper_access_parameters_variables,
            'wrapper_access_functions' : wrapper_access_functions,
            'wrapper_access_refractory' : wrapper_access_refractory,
            'wrapper_access_mean_fr' : wrapper_access_mean_fr,
            'wrapper_access_additional' : wrapper_access_additional,
        }

    @staticmethod
    def _pop_generate_default_wrapper(pop):
        """
        Generates the Python wrapper code for registered attributes.
        """
        datatypes = PyxGenerator._get_datatypes(pop)

        get_local_all = ""
        set_local_all = ""
        get_local = ""
        set_local = ""
        get_global = ""
        set_global = ""

        # Local parameters/variables
        for ctype in datatypes["local"]:
            ids = {
                'id': pop.id,
                'ctype': ctype,
                'ctype_name': ctype.replace(" ", "_")
            }

            # Population.get_cython_attribute expect np.array
            get_local_all += """
        if ctype == "%(ctype)s":
            return np.array(pop%(id)s.get_local_attribute_all_%(ctype_name)s(cpp_string))
""" % ids
            get_local += """
        if ctype == "%(ctype)s":
            return pop%(id)s.get_local_attribute_%(ctype_name)s(cpp_string, rk)
""" % ids

            # Setter
            set_local_all += """
        if ctype == "%(ctype)s":
            pop%(id)s.set_local_attribute_all_%(ctype_name)s(cpp_string, value)
""" % ids
            set_local += """
        if ctype == "%(ctype)s":
            pop%(id)s.set_local_attribute_%(ctype_name)s(cpp_string, rk, value)
""" % ids

        # Global parameters/variables
        for ctype in datatypes["global"]:
            ids = {
                'id': pop.id,
                'ctype': ctype,
                'ctype_name': ctype.replace(" ", "_")
            }

            get_global += """
        if ctype == "%(ctype)s":
            return pop%(id)s.get_global_attribute_%(ctype_name)s(cpp_string)
""" % ids

            set_global += """
        if ctype == "%(ctype)s":
            pop%(id)s.set_global_attribute_%(ctype_name)s(cpp_string, value)
""" % ids

        # Finalize code
        wrapper_code = ""

        if get_local_all != "":
            wrapper_code += PyxTemplate.pyx_default_pop_attribute_wrapper["local"] % {
                'get_local_all': get_local_all,
                'set_local_all': set_local_all,
                'get_local': get_local,
                'set_local': set_local
            }

        if get_global != "":
            wrapper_code += PyxTemplate.pyx_default_pop_attribute_wrapper["global"] % {
                'get_global': get_global,
                'set_global': set_global
            }
        
        return wrapper_code

#######################################################################
############## Projection #############################################
#######################################################################
    @staticmethod
    def _proj_struct(proj):
        """
        The python extension wrapper needs a definition of the corresponding
        C object. The pyx_struct contains all methods, which should be accessible
        by the python extension wrapper.

        Templates:

            structural_plasticity: pruning, creating, calling method
            delay, exact_integ: variables accessed by the wrapper

        """
        # Check for exact intgeration
        has_event_driven = False
        for var in proj.synapse_type.description['variables']:
            if var['method'] == 'event-driven':
                has_event_driven = True
                break

        # basic
        ids = {
            'id': proj.id,
            'float_prec': get_global_config('precision')
        }

        # Check if we need delay code
        has_delay = (proj.max_delay > 1)
        if proj.uniform_delay > 1 :
            key_delay = "uniform"
        else:
            if proj.synapse_type.type == "rate":
                key_delay = "nonuniform_rate_coded"
            else:
                key_delay = "nonuniform_spiking"

        # get the base templates
        template_dict = PyxGenerator._get_proj_template(proj)

        # Delay
        export_delay = ""
        if has_delay:
            if get_global_config('paradigm') == "openmp":
                export_delay = template_dict['delay'][key_delay]['pyx_struct'] % ids
            elif get_global_config('paradigm') == "cuda":
                export_delay = template_dict['delay'][key_delay]['pyx_struct'] % ids
            else:
                raise NotImplementedError

        # Event-driven
        export_event_driven = ""
        if has_event_driven:
            if get_global_config('paradigm') == "openmp":
                export_event_driven = template_dict['event_driven']['pyx_struct']
            elif get_global_config('paradigm') == "cuda":
                export_event_driven = template_dict['event_driven']['pyx_struct']
            else:
                raise NotImplementedError

        # Determine all export methods
        export_parameters_variables = ""

        # The transpose projection contains no own synaptic parameters
        if isinstance(proj, Transpose):
            export_parameters_variables = ""

        else:
            export_parameters_variables = ""
            datatypes = PyxGenerator._get_datatypes(proj)
            # Local parameters and variables
            for ctype in datatypes["local"]:
                export_parameters_variables += PyxTemplate.pyx_proj_attribute_export["local"] % {
                    'ctype': ctype,
                    'ctype_name': ctype.replace(" ", "_")
                }

            # Semiglobal parameters and variables
            for ctype in datatypes["semiglobal"]:
                export_parameters_variables += PyxTemplate.pyx_proj_attribute_export["semiglobal"] % {
                    'ctype': ctype,
                    'ctype_name': ctype.replace(" ", "_")
                }

            # Global parameters and variables
            for ctype in datatypes["global"]:
                export_parameters_variables += PyxTemplate.pyx_proj_attribute_export["global"] % {
                    'ctype': ctype,
                    'ctype_name': ctype.replace(" ", "_")
                }

        # Local functions
        export_functions, _ = PyxGenerator._custom_functions(proj)

        # Structural plasticity
        structural_plasticity = ""
        if get_global_config('structural_plasticity'):
            sp_tpl = template_dict['structural_plasticity']['pyx_struct']

            # Pruning in the synapse
            if 'pruning' in proj.synapse_type.description.keys():
                structural_plasticity += sp_tpl['pruning']
            if 'creating' in proj.synapse_type.description.keys():
                structural_plasticity += sp_tpl['creating']

            # Retrieve the names of extra attributes
            extra_args = ""
            for var in proj.synapse_type.description['parameters'] + proj.synapse_type.description['variables']:
                if not var['name'] in ['w', 'delay'] and  var['name'] in proj.synapse_type.description['local']:
                    extra_args += ', ' + var['ctype'] + ' ' +  var['name']
            # Generate the code
            structural_plasticity += sp_tpl['func'] % {'extra_args': extra_args}
            if proj.synapse_type.type == "spike":
                structural_plasticity += " "*8 + "void check_and_rebuild_inverse_connectivity()\n"

        # Check if either a custom definition or a CPP side init
        # is available otherwise fall back to init from LIL
        if proj.connector_name == "All-to-All" and cpp_connector_available("All-to-All", proj._storage_format, proj._storage_order):
            export_connector = tabify("bool all_to_all_pattern(vector[%(idx_type)s], vector[%(idx_type)s], %(float_prec)s, %(float_prec)s, %(float_prec)s, %(float_prec)s, bool)", 2)
        elif proj.connector_name == "Random" and cpp_connector_available("Random", proj._storage_format, proj._storage_order):
            export_connector = tabify("bool fixed_probability_pattern(vector[%(idx_type)s], vector[%(idx_type)s], %(float_prec)s, %(float_prec)s, %(float_prec)s, %(float_prec)s, %(float_prec)s, bool)", 2)
        elif proj.connector_name == "Random Convergent" and cpp_connector_available("Random Convergent", proj._storage_format, proj._storage_order):
            export_connector = tabify("bool fixed_number_pre_pattern(vector[%(idx_type)s], vector[%(idx_type)s], %(idx_type)s, %(float_prec)s, %(float_prec)s, %(float_prec)s, %(float_prec)s)", 2)
        else:
            export_connector = tabify("bool init_from_lil(vector[%(idx_type)s], vector[vector[%(idx_type)s]], vector[vector[%(float_prec)s]], vector[vector[int]], bool)", 2)

        # Data types, only of interest if "only_int_idx_type" configuration flag is false
        idx_types = determine_idx_type_for_projection(proj)
        idx_type_dict = {
            'float_prec': get_global_config('precision'),
            'idx_type': idx_types[1],
            'size_type': idx_types[3]
        }

        # Default LIL Definition/ Accessors
        # with an additional accessor spike
        default_conn_export = PyxTemplate.pyx_default_conn_export
        if proj.synapse_type.type == "spike":
            default_conn_export += """
        map[%(idx_type)s, %(idx_type)s] nb_efferent_synapses()
"""
        export_connector = export_connector % idx_type_dict
        export_connector_access = default_conn_export % idx_type_dict

        # Specific projections can overwrite
        if 'export_connector_call' in proj._specific_template.keys():
            export_connector = proj._specific_template['export_connector_call']
        if 'export_connectivity' in proj._specific_template.keys():
            export_connector_access = proj._specific_template['export_connectivity']
        if 'export_delay' in proj._specific_template.keys() and has_delay:
            export_delay = proj._specific_template['export_delay']
        if 'export_event_driven' in proj._specific_template.keys() and has_event_driven:
            export_event_driven = proj._specific_template['export_event_driven']
        if 'export_parameters_variables' in proj._specific_template.keys():
            export_parameters_variables = proj._specific_template['export_parameters_variables']

        # CUDA configuration update
        export_cuda_launch_config = ""
        if _check_paradigm("cuda"):
            export_cuda_launch_config = tabify("void update_launch_config(int, int)", 2)

        return PyxTemplate.proj_pyx_struct % {
            'id_proj': proj.id,
            'export_connectivity': export_connector+export_connector_access,
            'export_delay': export_delay,
            'export_event_driven': export_event_driven,
            'export_parameters_variables': export_parameters_variables,
            'export_functions': export_functions,
            'export_structural_plasticity': structural_plasticity,
            'export_additional': proj._specific_template['export_additional'] if 'export_additional' in proj._specific_template.keys() else "",
            'export_cuda_launch_config': export_cuda_launch_config
        }

    @staticmethod
    def _proj_wrapper(proj):
        """
        Generates the python extension wrapper, which allows access from Python
        to the C module. There are three optional parts (structural plasticity,
        non-uniform delays and exact integration of synaptic events) which we
        need to handle seperatly. The rest of the variables/parameters is handled
        by the standard accessors.

        Templates:

            attribute_pyx_wrapper: normal accessors for variables/parameters
            structural_plasticity: pruning, creating, calling method
            delay, exact_integ: __cinit__ code

        """
        # Check for exact intgeration
        has_event_driven = False
        for var in proj.synapse_type.description['variables']:
            if var['method'] == 'event-driven':
                has_event_driven = True
                break

        # basic
        ids = {
            'id_proj': proj.id,
            'float_prec': get_global_config('precision')
        }

        # Check if we need delay code
        has_delay = (proj.max_delay > 1)
        if proj.uniform_delay > 1 :
            key_delay = "uniform"
        else:
            if proj.synapse_type.type == "rate":
                key_delay = "nonuniform_rate_coded"
            else:
                key_delay = "nonuniform_spiking"

        # Import attributes templates
        wrapper_access_parameters_variables = PyxGenerator._proj_generate_default_export(proj)

        # select the base template
        template_dict = PyxGenerator._get_proj_template(proj)

        # Delays
        if not has_delay:
            wrapper_init_delay = ""
            wrapper_access_delay = ""
        else:
            # Initialize the wrapper
            wrapper_init_delay = template_dict['delay'][key_delay]['pyx_wrapper_init'] % ids
            # Access in wrapper
            wrapper_access_delay = template_dict['delay'][key_delay]['pyx_wrapper_accessor'] % ids

        # Local functions
        _, wrapper_access_functions = PyxGenerator._custom_functions(proj)


        # Additional declarations
        additional_declarations = ""

        # Structural plasticity
        structural_plasticity = ""
        if get_global_config('structural_plasticity'):
            if get_global_config('paradigm') == "openmp":
                sp_tpl = template_dict['structural_plasticity']['pyx_wrapper']
            else:
                sp_tpl = {}

            # Pruning in the synapse
            if 'pruning' in proj.synapse_type.description.keys():
                structural_plasticity += sp_tpl['pruning'] % {'id' : proj.id}

            # Creating in the synapse
            if 'creating' in proj.synapse_type.description.keys():
                structural_plasticity += sp_tpl['creating'] % {'id' : proj.id}

            # Retrieve the names of extra attributes
            extra_args = ""
            extra_values = ""
            for var in proj.synapse_type.description['parameters'] + proj.synapse_type.description['variables']:
                if not var['name'] in ['w', 'delay'] and  var['name'] in proj.synapse_type.description['local']:
                    extra_args += ', ' + var['ctype'] + ' ' +  var['name']
                    extra_values += ', ' +  var['name']

            # Generate the code
            structural_plasticity += sp_tpl['func'] % {'id' : proj.id, 'extra_args': extra_args, 'extra_values': extra_values}
            if proj.synapse_type.type == 'spike':
                structural_plasticity += """    def check_and_rebuild_inverse_connectivity(self):
        proj%(id_proj)s.check_and_rebuild_inverse_connectivity()
""" % {'id_proj': proj.id}

        # Check if either a custom definition or a CPP side init
        # is available otherwise fall back to init from LIL
        if proj.connector_name == "All-to-All" and cpp_connector_available("All-to-All", proj._storage_format, proj._storage_order):
            wrapper_connector_call = """
    def all_to_all(self, post_ranks, pre_ranks, w_dist_arg1, w_dist_arg2, d_dist_arg1, d_dist_arg2, allow_self_connections):
        return proj%(id_proj)s.all_to_all_pattern(post_ranks, pre_ranks, w_dist_arg1, w_dist_arg2, d_dist_arg1, d_dist_arg2, allow_self_connections)
""" % {'id_proj': proj.id}
        elif proj.connector_name == "Random" and cpp_connector_available("Random", proj._storage_format, proj._storage_order):
            wrapper_connector_call = """
    def fixed_probability(self, post_ranks, pre_ranks, p, w_dist_arg1, w_dist_arg2, d_dist_arg1, d_dist_arg2, allow_self_connections):
        return proj%(id_proj)s.fixed_probability_pattern(post_ranks, pre_ranks, p, w_dist_arg1, w_dist_arg2, d_dist_arg1, d_dist_arg2, allow_self_connections)
""" % {'id_proj': proj.id}
        elif proj.connector_name == "Random Convergent" and cpp_connector_available("Random Convergent", proj._storage_format, proj._storage_order):
            wrapper_connector_call = """
    def fixed_number_pre(self, post_ranks, pre_ranks, number_synapses_per_row, w_dist_arg1, w_dist_arg2, d_dist_arg1, d_dist_arg2):
        return proj%(id_proj)s.fixed_number_pre_pattern(post_ranks, pre_ranks, number_synapses_per_row, w_dist_arg1, w_dist_arg2, d_dist_arg1, d_dist_arg2)
""" % {'id_proj': proj.id}
        else:
            wrapper_connector_call = """
    def init_from_lil_connectivity(self, synapses):
        " synapses is an instance of LILConnectivity "
        return proj%(id_proj)s.init_from_lil(synapses.post_rank, synapses.pre_rank, synapses.w, synapses.delay, synapses.requires_sorting)

    def init_from_lil(self, post_rank, pre_rank, w, delay, requires_sorting):
        return proj%(id_proj)s.init_from_lil(post_rank, pre_rank, w, delay, requires_sorting)
""" % {'id_proj': proj.id}

        wrapper_args = ""
        wrapper_init = tabify("pass",3)
        wrapper_access_connectivity = PyxTemplate.pyx_default_conn_wrapper 
        if proj.synapse_type.type == "spike":   # additional for spike
            wrapper_access_connectivity += """
    def nb_efferent_synapses(self):
        return proj%(id_proj)s.nb_efferent_synapses()
"""
        wrapper_access_connectivity = wrapper_access_connectivity % {'id_proj': proj.id}

        # Specific projections can overwrite
        if 'wrapper_args' in proj._specific_template.keys():
            wrapper_args = proj._specific_template['wrapper_args']
        if 'wrapper_init_connectivity' in proj._specific_template.keys():
            wrapper_init = proj._specific_template['wrapper_init_connectivity']
        if 'wrapper_access_connectivity' in proj._specific_template.keys():
            wrapper_access_connectivity = proj._specific_template['wrapper_access_connectivity']
        if 'wrapper_connector_call' in proj._specific_template.keys():
            wrapper_connector_call = proj._specific_template['wrapper_connector_call']
        if 'wrapper_init_delay' in proj._specific_template.keys() and has_delay:
            wrapper_init_delay = proj._specific_template['wrapper_init_delay']
        if 'wrapper_access_delay' in proj._specific_template.keys() and has_delay:
            wrapper_access_delay = proj._specific_template['wrapper_access_delay']
        if 'wrapper_access_parameters_variables' in proj._specific_template.keys():
            wrapper_access_parameters_variables = proj._specific_template['wrapper_access_parameters_variables']
        if 'wrapper_access_additional' in proj._specific_template.keys():
            additional_declarations = proj._specific_template['wrapper_access_additional']

        # CUDA configuration update
        wrapper_cuda_launch_config = ""
        if _check_paradigm("cuda"):
            wrapper_cuda_launch_config = """
    def update_launch_config(self, nb_blocks=-1, threads_per_block=32):
        proj%(id_proj)s.update_launch_config(nb_blocks, threads_per_block)
""" % {'id_proj': proj.id}

        return PyxTemplate.proj_pyx_wrapper % {
            'id_proj': proj.id,
            'pre_size': proj.pre.population.size if isinstance(proj.pre, PopulationView) else proj.pre.size,
            'post_size': proj.post.population.size if isinstance(proj.post, PopulationView) else proj.post.size,
            'wrapper_args' : wrapper_args,
            'wrapper_init' : wrapper_init,
            'wrapper_connector_call': wrapper_connector_call,
            'wrapper_init_delay': wrapper_init_delay,
            'wrapper_access_connectivity': wrapper_access_connectivity,
            'wrapper_access_delay': wrapper_access_delay,
            'wrapper_access_parameters_variables': wrapper_access_parameters_variables,
            'wrapper_access_functions': wrapper_access_functions,
            'wrapper_access_structural_plasticity': structural_plasticity,
            'wrapper_access_additional': additional_declarations,
            'wrapper_cuda_launch_config': wrapper_cuda_launch_config
        }

    @staticmethod
    def _proj_generate_default_export(proj):
        """
        To prevent overloaded functions by different types, we need to declare
        all accessors per c++ data type which is used.
        """
        # The transpose projection contains no own synaptic parameters
        if isinstance(proj, Transpose):
            return ""

        get_local_all = ""
        set_local_all = ""
        get_local_row = ""
        set_local_row = ""
        get_local = ""
        set_local = ""
        get_semiglobal_all = ""
        set_semiglobal_all = ""
        get_semiglobal = ""
        set_semiglobal = ""
        set_global = ""
        get_global = ""

        datatypes = PyxGenerator._get_datatypes(proj)
        for ctype in datatypes["local"]:
            ids = {
                'id_proj': proj.id,
                'ctype': ctype,
                'ctype_name': ctype.replace(" ", "_")
            }

            get_local_all += """
        if ctype == "%(ctype)s":
            return proj%(id_proj)s.get_local_attribute_all_%(ctype_name)s(cpp_string)
""" % ids
            set_local_all += """
        if ctype == "%(ctype)s":
            proj%(id_proj)s.set_local_attribute_all_%(ctype_name)s(cpp_string, value)
""" % ids
            get_local_row += """
        if ctype == "%(ctype)s":
            return proj%(id_proj)s.get_local_attribute_row_%(ctype_name)s(cpp_string, rk_post)
""" % ids
            set_local_row += """
        if ctype == "%(ctype)s":
            proj%(id_proj)s.set_local_attribute_row_%(ctype_name)s(cpp_string, rk_post, value)
""" % ids
            get_local += """
        if ctype == "%(ctype)s":
            return proj%(id_proj)s.get_local_attribute_%(ctype_name)s(cpp_string, rk_post, rk_pre)
""" % ids
            set_local += """
        if ctype == "%(ctype)s":
            proj%(id_proj)s.set_local_attribute_%(ctype_name)s(cpp_string, rk_post, rk_pre, value)
""" % ids

        for ctype in datatypes["semiglobal"]:
            ids = {
                'id_proj': proj.id,
                'ctype': ctype,
                'ctype_name': ctype.replace(" ", "_")
            }

            get_semiglobal_all += """
        if ctype == "%(ctype)s":
            return proj%(id_proj)s.get_semiglobal_attribute_all_%(ctype_name)s(cpp_string)
""" % ids
            set_semiglobal_all += """
        if ctype == "%(ctype)s":
            proj%(id_proj)s.set_semiglobal_attribute_all_%(ctype_name)s(cpp_string, value)
""" % ids
            get_semiglobal += """
        if ctype == "%(ctype)s":
            return proj%(id_proj)s.get_semiglobal_attribute_%(ctype_name)s(cpp_string, rk_post)
""" % ids
            set_semiglobal += """
        if ctype == "%(ctype)s":
            proj%(id_proj)s.set_semiglobal_attribute_%(ctype_name)s(cpp_string, rk_post, value)
""" % ids

        for ctype in datatypes["global"]:
            ids = {
                'id_proj': proj.id,
                'ctype': ctype,
                'ctype_name': ctype.replace(" ", "_")
            }

            get_global += """
        if ctype == "%(ctype)s":            
            return proj%(id_proj)s.get_global_attribute_%(ctype_name)s(cpp_string)
""" % ids
            set_global += """
        if ctype == "%(ctype)s":            
            proj%(id_proj)s.set_global_attribute_%(ctype_name)s(cpp_string, value)
""" % ids

        # Finalize code
        wrapper_code = ""

        if get_local_all != "":
            wrapper_code += PyxTemplate.pyx_proj_attribute_wrapper["local"] % {
                'get_local_all': get_local_all,
                'set_local_all': set_local_all,
                'get_local_row': get_local_row,
                'set_local_row': set_local_row,
                'get_local': get_local,
                'set_local': set_local,
                'id_proj': proj.id
            }

        if get_semiglobal_all != "":
            wrapper_code += PyxTemplate.pyx_proj_attribute_wrapper["semiglobal"] % {
                'get_semiglobal_all': get_semiglobal_all,
                'set_semiglobal_all': set_semiglobal_all,
                'get_semiglobal': get_semiglobal,
                'set_semiglobal': set_semiglobal,
                'id_proj': proj.id
            }

        if get_global != "":
            wrapper_code += PyxTemplate.pyx_proj_attribute_wrapper["global"] % {
                'get_global': get_global,
                'set_global': set_global,
                'id_proj': proj.id
            }

        return wrapper_code

#######################################################################
############## Monitors  ##############################################
#######################################################################
    @staticmethod
    def _pop_monitor_struct(pop):
        """
        Generate recorder struct.
        """
        tpl_code = """
    # Population %(id)s (%(name)s) : Monitor
    cdef cppclass PopRecorder%(id)s (Monitor):
        @staticmethod
        int create_instance(vector[int], int, int, long)
        @staticmethod
        PopRecorder%(id)s* get_instance(int)
        long int size_in_bytes()
        void clear()
"""
        attributes = []
        for var in pop.neuron_type.description['variables']:
            # Avoid doublons
            if var['name'] in attributes:
                continue
            attributes.append(var['name'])

            if var['name'] in pop.neuron_type.description['local']:
                tpl_code += """
        vector[vector[%(type)s]] %(name)s
        bool record_%(name)s
""" % {'name': var['name'], 'type': var['ctype']}
            elif var['name'] in pop.neuron_type.description['global']:
                tpl_code += """
        vector[%(type)s] %(name)s
        bool record_%(name)s
""" % {'name': var['name'], 'type': var['ctype']}

        if pop.neuron_type.type == 'spike':
            tpl_code += """
        map[int, vector[long]] spike
        bool record_spike
        void clear_spike()
"""
            if pop.neuron_type.axon_spike:
                tpl_code += """
        map[int, vector[long]] axon_spike
        bool record_axon_spike
        void clear_axon_spike()
"""

        # Arrays for the presynaptic sums
        if pop.neuron_type.type == 'rate':
            tpl_code += """
        # Targets"""
            for target in sorted(list(set(pop.neuron_type.description['targets'] + pop.targets))):
                tpl_code += """
        vector[vector[%(float_prec)s]] _sum_%(target)s
        bool record__sum_%(target)s
""" % {'target': target, 'float_prec': get_global_config('precision')}

        return tpl_code % {'id' : pop.id, 'name': pop.name}

    @staticmethod
    def _pop_monitor_wrapper(pop):
        """
        Generate recorder wrapper.
        """
        tpl_code = """
# Population Monitor wrapper
@cython.auto_pickle(True)
cdef class PopRecorder%(id)s_wrapper:
    cdef int id
    def __init__(self, list ranks, int period, period_offset, long offset):
        self.id = PopRecorder%(id)s.create_instance(ranks, period, period_offset, offset)

    def size_in_bytes(self):
        return (PopRecorder%(id)s.get_instance(self.id)).size_in_bytes()

    def clear(self):
        return (PopRecorder%(id)s.get_instance(self.id)).clear()

    property period:
        def __get__(self): return (PopRecorder%(id)s.get_instance(self.id)).period_
        def __set__(self, val): (PopRecorder%(id)s.get_instance(self.id)).period_ = val

    property period_offset:
        def __get__(self): return (PopRecorder%(id)s.get_instance(self.id)).period_offset_
        def __set__(self, val): (PopRecorder%(id)s.get_instance(self.id)).period_offset_ = val
"""
        attributes = []
        for var in pop.neuron_type.description['variables']:
            # Avoid doublons
            if var['name'] in attributes:
                continue
            attributes.append(var['name'])
            tpl_code += """
    property %(name)s:
        def __get__(self): return (PopRecorder%(id)s.get_instance(self.id)).%(name)s
        def __set__(self, val): (PopRecorder%(id)s.get_instance(self.id)).%(name)s = val
    property record_%(name)s:
        def __get__(self): return (PopRecorder%(id)s.get_instance(self.id)).record_%(name)s
        def __set__(self, val): (PopRecorder%(id)s.get_instance(self.id)).record_%(name)s = val
    def clear_%(name)s(self):
        (PopRecorder%(id)s.get_instance(self.id)).%(name)s.clear()
""" % {'id' : pop.id, 'name': var['name']}

        if pop.neuron_type.type == 'spike':
            tpl_code += """
    property spike:
        def __get__(self): return (PopRecorder%(id)s.get_instance(self.id)).spike
        def __set__(self, val): (PopRecorder%(id)s.get_instance(self.id)).spike = val
    property record_spike:
        def __get__(self): return (PopRecorder%(id)s.get_instance(self.id)).record_spike
        def __set__(self, val): (PopRecorder%(id)s.get_instance(self.id)).record_spike = val
    def clear_spike(self):
        (PopRecorder%(id)s.get_instance(self.id)).clear_spike()
""" % {'id' : pop.id}

            if pop.neuron_type.axon_spike:
                tpl_code += """
    property axon_spike:
        def __get__(self): return (PopRecorder%(id)s.get_instance(self.id)).axon_spike
        def __set__(self, val): (PopRecorder%(id)s.get_instance(self.id)).axon_spike = val
    property record_axon_spike:
        def __get__(self): return (PopRecorder%(id)s.get_instance(self.id)).record_axon_spike
        def __set__(self, val): (PopRecorder%(id)s.get_instance(self.id)).record_axon_spike = val
    def clear_axon_spike(self):
        (PopRecorder%(id)s.get_instance(self.id)).clear_axon_spike()
""" % {'id' : pop.id}

        # Arrays for the presynaptic sums
        if pop.neuron_type.type == 'rate':
            tpl_code += """
    # Targets"""
            for target in sorted(list(set(pop.neuron_type.description['targets'] + pop.targets))):
                tpl_code += """
    property %(name)s:
        def __get__(self): return (PopRecorder%(id)s.get_instance(self.id)).%(name)s
        def __set__(self, val): (PopRecorder%(id)s.get_instance(self.id)).%(name)s = val
    property record_%(name)s:
        def __get__(self): return (PopRecorder%(id)s.get_instance(self.id)).record_%(name)s
        def __set__(self, val): (PopRecorder%(id)s.get_instance(self.id)).record_%(name)s = val
    def clear_%(name)s(self):
        (PopRecorder%(id)s.get_instance(self.id)).%(name)s.clear()
""" % {'id' : pop.id, 'name': '_sum_'+target}

        return tpl_code % {'id' : pop.id, 'name': pop.name}

    @staticmethod
    def _proj_monitor_struct(proj):
        """
        Generate projection recorder struct
        """

        # Specific template
        if 'monitor_export' in proj._specific_template.keys():
            return proj._specific_template['monitor_export']


        code = """
    # Projection %(id)s : Monitor
    cdef cppclass ProjRecorder%(id)s (Monitor):
        @staticmethod
        int create_instance(vector[int], int, int, long)
        @staticmethod
        ProjRecorder%(id)s* get_instance(int)
"""

        templates = {
        'local': """
        vector[vector[vector[%(type)s]]] %(name)s
        bool record_%(name)s
""",
        'semiglobal': """
        vector[vector[%(type)s]] %(name)s
        bool record_%(name)s
""",
        'global': """
        vector[%(type)s] %(name)s
        bool record_%(name)s
"""
        }

        attributes = []
        for var in proj.synapse_type.description['variables']:
            # Avoid doublons
            if var['name'] in attributes:
                continue
            attributes.append(var['name'])

            # Get the locality
            locality = var['locality']

            # Special case for single weights
            if var['name'] == "w" and proj._has_single_weight():
                locality = 'global'

            # Use the correct template
            code +=  templates[locality] % {'name': var['name'], 'type': var['ctype']}

        return code % {'id' : proj.id}

    @staticmethod
    def _proj_monitor_wrapper(proj):
        """
        Generate projection recorder struct
        """

        # Specific template
        if 'monitor_wrapper' in proj._specific_template.keys():
            return proj._specific_template['monitor_wrapper']

        code = """
# Projection Monitor wrapper
@cython.auto_pickle(True)
cdef class ProjRecorder%(id)s_wrapper:
    cdef int id
    def __init__(self, list ranks, int period, int period_offset, long offset):
        self.id = ProjRecorder%(id)s.create_instance(ranks, period, period_offset, offset)
"""

        attributes = []
        for var in proj.synapse_type.description['variables']:
            # Avoid doublons
            if var['name'] in attributes:
                continue
            attributes.append(var['name'])

            code += """
    property %(name)s:
        def __get__(self): return (ProjRecorder%(id)s.get_instance(self.id)).%(name)s
        def __set__(self, val): (ProjRecorder%(id)s.get_instance(self.id)).%(name)s = val
    property record_%(name)s:
        def __get__(self): return (ProjRecorder%(id)s.get_instance(self.id)).record_%(name)s
        def __set__(self, val): (ProjRecorder%(id)s.get_instance(self.id)).record_%(name)s = val
    def clear_%(name)s(self):
        (ProjRecorder%(id)s.get_instance(self.id)).%(name)s.clear()
""" % {'id' : proj.id, 'name': var['name']}

        return code % {'id' : proj.id}

#######################################################################
############## Helpers   ##############################################
#######################################################################
    @staticmethod
    def _get_datatypes(obj):
        """
        Helper method used in proj_wrapper/proj_export or pop_wrapper/pop_export
        """
        if isinstance(obj, Projection):
            datatypes = {
                'local': [],
                'semiglobal': [],
                'global': []
            }

            for var in obj.synapse_type.description['parameters'] + obj.synapse_type.description['variables']:
                locality = var['locality']
                if var['name'] == "w" and obj._has_single_weight():
                    locality = 'global'

                if var['ctype'] not in datatypes[locality]:
                    datatypes[locality].append(var['ctype'])

        elif isinstance(obj, Population):
            datatypes = {
                'local': [],
                'global': []
            }

            if obj.neuron_type.type == 'spike':
                if 'int' not in datatypes['local']:
                    datatypes['local'].append('int')

            for var in obj.neuron_type.description['parameters'] + obj.neuron_type.description['variables']:
                locality = var['locality']
                if var['ctype'] not in datatypes[locality]:
                    datatypes[locality].append(var['ctype'])

        else:
            ValueError("PyxGenerator._get_datatypes() expects either Population or Projection instance.")

        return datatypes



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

# Code which should be added prior to kernels
# (directly imported by CodeGenerator if needed)
additional_global_functions = ""

launch_config = {
    'init': """
        _threads_per_block = 64;
        _nb_blocks = std::min<unsigned int>(nb_dendrites(), 65535);
    
    #ifdef _DEBUG
        std::cout << "Kernel configuration: " << _nb_blocks << ", " << _threads_per_block << std::endl;
    #endif
""",
    'update': """
        if (nb_blocks != -1) {
            _nb_blocks = static_cast<unsigned int>(nb_blocks);
            _threads_per_block = threads_per_block;
        }else{
            _threads_per_block = threads_per_block;
            _nb_blocks = std::min<unsigned int>(nb_dendrites(), 65535);
        }

    #ifdef _DEBUG
        std::cout << "Updated configuration: " << _nb_blocks << ", " << _threads_per_block << std::endl;
    #endif
"""
}

attribute_decl = {
    'local': """
    // Local %(attr_type)s %(name)s
    std::vector< %(type)s > %(name)s;
    %(type)s* gpu_%(name)s;
    long int %(name)s_device_to_host;
    bool %(name)s_host_to_device;
""",
    'semiglobal': """
    // Semiglobal %(attr_type)s %(name)s
    std::vector< %(type)s >  %(name)s ;
    %(type)s* gpu_%(name)s;
    long int %(name)s_device_to_host;
    bool %(name)s_host_to_device;
""",
    'global': {
        'parameter': """
    // Global %(attr_type)s %(name)s
    %(type)s %(name)s;
""",
        'variable': """
    // Global %(attr_type)s %(name)s
    %(type)s %(name)s;
    %(type)s* gpu_%(name)s;
    long int %(name)s_device_to_host;
    bool %(name)s_host_to_device;
"""
    }
}

attribute_cpp_init = {
    'local': """
        // Local %(attr_type)s %(name)s
        %(name)s = init_matrix_variable<%(type)s>(%(init)s);
        gpu_%(name)s = init_matrix_variable_gpu<%(type)s>(%(name)s);
        %(name)s_host_to_device = true;
        %(name)s_device_to_host = t;
""",
    'semiglobal': """
        // Semiglobal %(attr_type)s %(name)s
        %(name)s = init_vector_variable<%(type)s>(%(init)s);
        gpu_%(name)s = init_vector_variable_gpu<%(type)s>(%(name)s);
        %(name)s_host_to_device = true;
        %(name)s_device_to_host = t;
""",
    'global': {
        'parameter': """
        // Global %(attr_type)s %(name)s
        %(name)s = 0.0;
    """,
        'variable': """
        // Global %(attr_type)s %(name)s
        %(name)s = static_cast<%(type)s>(%(init)s);
        cudaMalloc((void**)&gpu_%(name)s, sizeof(%(type)s));
        %(name)s_host_to_device = true;
        %(name)s_device_to_host = t;
"""
    }
}

attribute_cpp_size = {
    'local': """
        // Local %(attr_type)s %(name)s
        size_in_bytes += sizeof(bool);
        size_in_bytes += sizeof(%(ctype)s*);
        size_in_bytes += sizeof(std::vector<%(ctype)s>);
        size_in_bytes += sizeof(%(ctype)s) * %(name)s.capacity();
""",
    'semiglobal': """
        // Semiglobal %(attr_type)s %(name)s
        size_in_bytes += sizeof(bool);
        size_in_bytes += sizeof(%(ctype)s*);
        size_in_bytes += sizeof(std::vector<%(ctype)s>);
        size_in_bytes += sizeof(%(ctype)s) * %(name)s.capacity();
""",
    'global': """
        // Global
        size_in_bytes += sizeof(bool);
        size_in_bytes += sizeof(%(ctype)s*);
        size_in_bytes += sizeof(%(ctype)s);
"""
}

attribute_cpp_delete = {
    'local': """
        // %(name)s - host
        %(name)s.clear();
        %(name)s.shrink_to_fit();

        // %(name)s - device
        cudaFree(gpu_%(name)s);
""",
    'semiglobal': """
        // %(name)s - host
        %(name)s.clear();
        %(name)s.shrink_to_fit();

        // %(name)s - device
        cudaFree(gpu_%(name)s);
""",
    'global': ""
}

attribute_host_to_device = {
    'local': """
        // %(name)s: local
        if ( %(name)s_host_to_device )
        {
        #ifdef _DEBUG
            std::cout << "HtoD: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            cudaMemcpy( gpu_%(name)s, %(name)s.data(), num_non_zeros_ * sizeof( %(type)s ), cudaMemcpyHostToDevice);
            %(name)s_host_to_device = false;
        #ifdef _DEBUG
            cudaError_t err = cudaGetLastError();
            if ( err!= cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err) << std::endl;
        #endif
        }
""",
    'semiglobal': """
        // %(name)s: semiglobal
        if ( %(name)s_host_to_device )
        {
        #ifdef _DEBUG
            std::cout << "HtoD: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            cudaMemcpy( gpu_%(name)s, %(name)s.data(), post_ranks_.size() * sizeof( %(type)s ), cudaMemcpyHostToDevice);
            %(name)s_host_to_device = false;
        #ifdef _DEBUG
            cudaError_t err = cudaGetLastError();
            if ( err!= cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err) << std::endl;
        #endif
        }
""",
    'global': """
        // %(name)s: global
        if ( %(name)s_host_to_device )
        {
        #ifdef _DEBUG
            std::cout << "HtoD: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            cudaMemcpy( gpu_%(name)s, &%(name)s, sizeof( %(type)s ), cudaMemcpyHostToDevice);
            %(name)s_host_to_device = false;
        #ifdef _DEBUG
            cudaError_t err = cudaGetLastError();
            if ( err!= cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err) << std::endl;
        #endif
        }
"""
}

attribute_device_to_host = {
    'local': """
        // %(name)s: local
        if ( %(name)s_device_to_host < t ) {
        #ifdef _DEBUG
            std::cout << "DtoH: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            cudaMemcpy( %(name)s.data(), gpu_%(name)s, num_non_zeros_ * sizeof( %(type)s ), cudaMemcpyDeviceToHost);
        #ifdef _DEBUG
            cudaError_t err_%(name)s = cudaGetLastError();
            if ( err_%(name)s != cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err_%(name)s) << std::endl;
        #endif
            %(name)s_device_to_host = t;
        }
""",
    'semiglobal': """
            // %(name)s: semiglobal
        #ifdef _DEBUG
            std::cout << "DtoH: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            cudaMemcpy( %(name)s.data(), gpu_%(name)s, post_ranks_.size() * sizeof(%(type)s), cudaMemcpyDeviceToHost);
        #ifdef _DEBUG
            cudaError_t err_%(name)s = cudaGetLastError();
            if ( err_%(name)s != cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err_%(name)s) << std::endl;
        #endif
""",
    'global': """
            // %(name)s: global
        #ifdef _DEBUG
            std::cout << "DtoH: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            cudaMemcpy( &%(name)s, gpu_%(name)s, sizeof(%(type)s), cudaMemcpyDeviceToHost);
        #ifdef _DEBUG
            cudaError_t err_%(name)s = cudaGetLastError();
            if ( err_%(name)s != cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err_%(name)s) << std::endl;
        #endif
"""
}

delay = {
    'uniform': {
        'declare': """
    // Uniform delay
    int delay ;""",
        'pyx_struct': """
        # Non-uniform delay
        int delay""",
        'init': "delay = delays[0][0];",
        'pyx_wrapper_init': """
        proj%(id_proj)s.delay = syn.uniform_delay""",
        'pyx_wrapper_accessor': """
    # Access to non-uniform delay
    def get_delay(self):
        return proj%(id_proj)s.delay
    def get_dendrite_delay(self, idx):
        return proj%(id_proj)s.delay
    def set_delay(self, value):
        print("set delay", value)
        proj%(id_proj)s.delay = value
"""
    },
    'nonuniform_rate_coded': {
        'declare': """
    // Non-uniform delay
    std::vector< std::vector< int > > delay ;""",
        'pyx_struct': """
        # Non-uniform delay
        vector[vector[int]] delay""",
    
        'init': "",

        'pyx_wrapper_init': """
        proj%(id_proj)s.delay = syn.delay""",
    
        'pyx_wrapper_accessor': """
    # Access to non-uniform delay
    def get_delay(self):
        return proj%(id_proj)s.delay
    def get_dendrite_delay(self, idx):
        return proj%(id_proj)s.delay[idx]
    def set_delay(self, value):
        proj%(id_proj)s.delay = value
"""
    },
    'nonuniform_spiking': None
}

event_driven = {
    'declare': """
    std::vector< long > _last_event;
    long* _gpu_last_event;
""",
    'cpp_init': """
    _last_event = init_matrix_variable<long>(-10000);
    _gpu_last_event = init_matrix_variable_gpu<long>(_last_event);    
""",
    'pyx_struct': """
        vector[long] _last_event
""",
    'pyx_wrapper_init':
"""
        proj%(id_proj)s._last_event = vector[long]( syn._matrix.num_elements(), -10000)
"""
}

spike_event_transmission = {
    'device_kernel': """// gpu device kernel for projection %(id_proj)s
__global__ void cu_proj%(id_proj)s_psp( %(float_prec)s dt, bool plasticity, int *spiked, unsigned int* spike_count, %(conn_args_header)s %(kernel_args_header)s ) {
    int b_idx = blockIdx.x;
    int pre_rank = spiked[b_idx];

    while (b_idx < *spike_count) {
        int syn_idx = threadIdx.x + row_ptr[pre_rank];
        while (syn_idx < row_ptr[pre_rank+1]){
%(psp)s
            syn_idx += blockDim.x;
        }
        b_idx += gridDim.x;
    }

}
""",
    'invoke_kernel': """
void proj%(id_proj)s_psp(RunConfig cfg, %(float_prec)s dt, bool plasticity, int *spiked, unsigned int* spike_count,  %(conn_args_header)s %(kernel_args_header)s) {
    cu_proj%(id_proj)s_psp<<< cfg.nb, cfg.tpb, cfg.smem_size, cfg.stream >>>(
        /* default events */
        dt, plasticity,
        /* pre-synaptic events */
        spiked, spike_count,
        /* connectivity */
        %(conn_args_invoke)s
        /* other arguments */
        %(kernel_args_invoke)s
    );
}
""",
    'kernel_decl': """void proj%(id_proj)s_psp(RunConfig cfg, %(float_prec)s dt, bool plasticity, int *spiked, unsigned int* spike_count,  %(conn_args_header)s %(kernel_args_header)s);
""",
    'host_call': """
        if ( pop%(id_pre)s._active && (pop%(id_pre)s.spike_count > 0) && proj%(id_proj)s._transmission) {
            int tpb = 1024;//__pop%(id_pre)s_pop%(id_post)s_%(target)s_tpb__;
            int nbBlocks = pop%(id_pre)s.spike_count;
            
            proj%(id_proj)s_psp(
                RunConfig(nbBlocks, tpb, 0, proj%(id_proj)s.stream),
                /* default events */
                dt, proj%(id_proj)s._plasticity,
                /* pre-synaptic events */
                pop%(id_pre)s.gpu_spiked, pop%(id_pre)s.gpu_spike_count,
                %(conn_args)s
                /* other arguments */
                %(kernel_args)s
            );

        #ifdef _DEBUG
            cudaDeviceSynchronize();
            cudaError_t err_psp_proj%(id_proj)s = cudaGetLastError();
            if( err_psp_proj%(id_proj)s != cudaSuccess) {
                std::cout << "proj%(id_proj)s_psp (" << t << "): " << std::endl;
                std::cout << "   " << cudaGetErrorString(err_psp_proj%(id_proj)s) << std::endl;
            }
        #endif
        }
"""
}

conn_templates = {
    # connectivity representation
    'conn_header' : "const %(idx_type)s post_size, const %(idx_type)s* __restrict__  rank_post, const %(size_type)s* __restrict__ row_ptr, const %(idx_type)s* __restrict__ rank_pre",
    'conn_call' : "proj%(id_proj)s.nb_dendrites(), proj%(id_proj)s.gpu_post_rank, proj%(id_proj)s.gpu_row_ptr, proj%(id_proj)s.gpu_pre_rank",
    'conn_kernel': "",

    # launch config
    'launch_config': launch_config,

    # accessors
    'attribute_decl': attribute_decl,
    'attribute_cpp_init': attribute_cpp_init,
    'attribute_cpp_size': attribute_cpp_size,
    'attribute_cpp_delete':attribute_cpp_delete,
    'host_to_device': attribute_host_to_device,
    'device_to_host': attribute_device_to_host,
    'delay': delay,
    'event_driven': event_driven,

    # operations
    'spike_transmission': {
        'event_driven': spike_event_transmission
    }
}

conn_ids = {
    'local_index': "[j]",
    'semiglobal_index': '[i]',
    'global_index': '[0]',
    'pre_index': '[rank_pre[j]]',
    'post_index': '[col_idx[syn_idx]]',
    'pre_prefix': 'pre_',
    'post_prefix': 'post_',
    'delay_nu' : '[delay[j]-1]', # non-uniform delay
}


"""

    test_GPU.py

    This file is part of ANNarchy.

    Copyright (C) 2013-2016 Joseph Gussev <joseph.gussev@s2012.tu-chemnitz.de>,
    Helge Uelo Dinkelbach <helge.dinkelbach@gmail.com>,
    Julien Vitay <julien.vitay@informatik.tu-chemnitz.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    ANNarchy is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
import unittest
from ANNarchy import setup

setup(paradigm="cuda")

from Common import *
from Interface import *
from Neuron import *
from Synapse import *


if __name__ == '__main__':
    unittest.main(verbosity=2)



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

import numpy as np

from ANNarchy.intern.SpecificPopulation import SpecificPopulation
from ANNarchy.intern.ConfigManagement import get_global_config
from ANNarchy.intern import Messages
from ANNarchy.core.Population import Population
from ANNarchy.core.Neuron import Neuron

class TimedArray(SpecificPopulation):
    """
    Data structure holding sequential inputs for a rate-coded network.

    The input values are stored in the (recordable) attribute `r`, without any further processing.
    You will need to connect this population to another one using the ``connect_one_to_one()`` method.

    By default, the firing rate of this population will iterate over the different values step by step:

    ```python
    inputs = np.array(
        [
            [1, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 1, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 1, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 1, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 1, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 1, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 1, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 1]
        ]
    )

    inp = ann.TimedArray(rates=inputs)

    pop = ann.Population(10, ...)

    proj = ann.Projection(inp, pop, 'exc')
    proj.connect_one_to_one(1.0)

    ann.compile()

    ann.simulate(10.)
    ```

    This creates a population of 10 neurons whose activity will change during the first 10*dt milliseconds of the simulation. After that delay, the last input will be kept (i.e. 1 for the last neuron).

    If you want the TimedArray to "loop" over the different input vectors, you can specify a period for the inputs:

    ```python
    inp = ann.TimedArray(rates=inputs, period=10.)
    ```

    If the period is smaller than the length of the rates, the last inputs will not be set.

    If you do not want the inputs to be set at every step, but every 10 ms for example, youcan use the ``schedule`` argument:

    ```python
    inp = ann.TimedArray(rates=inputs, schedule=10.)
    ```

    The input [1, 0, 0,...] will stay for 10 ms, then[0, 1, 0, ...] for the next 10 ms, etc...

    If you need a less regular schedule, you can specify it as a list of times:

    ```python
    inp = ann.TimedArray(rates=inputs, schedule=[10., 20., 50., 60., 100., 110.])
    ```

    The first input is set at t = 10 ms (r = 0.0 in the first 10 ms), the second at t = 20 ms, the third at t = 50 ms, etc.

    If you specify less times than in the array of rates, the last ones will be ignored.

    Scheduling can be combined with periodic cycling. Note that you can use the ``reset()`` method to manually reinitialize the TimedArray, times becoming relative to that call:

    ```python
    ann.simulate(100.) # ten inputs are shown with a schedule of 10 ms
    inp.reset()
    ann.simulate(100.) # the same ten inputs are presented again.
    ```

    :param rates: array of firing rates. The first axis corresponds to time, the others to the desired dimensions of the population.
    :param geometry: desired dimensions of the population. This argument will be considered if *rates* is None.
    :param schedule: either a single value or a list of time points where inputs should be set. Default: every timestep.
    :param period: time when the timed array will be reset and start again, allowing cycling over the inputs. Default: no cycling (-1.).

    """
    def __init__(self, 
                 rates:np.ndarray=None, 
                 geometry:int|tuple=None, 
                 schedule:float=0., 
                 period:float=-1., 
                 name:str=None, 
                 copied:bool=False):

        neuron = Neuron(
            parameters="",
            equations="r = 0.0",
            name="Timed Array",
            description="Timed array source."
        )

        # Sanity check
        if rates is None and geometry is None:
            Messages._error("TimedArray: either *rates* or *geometry* argument must be set.")

        # Geometry of the population
        if rates is not None:
            if geometry is None:
                geometry = rates.shape[1:]
            else:
                if geometry != rates.shape[1:]:
                    Messages._warning("TimedArray: mismatch between *rates* and *geometry* dimensions detected.")

        SpecificPopulation.__init__(self, geometry=geometry, neuron=neuron, name=name, copied=copied)

        self.init['schedule'] = schedule
        self.init['rates'] = rates
        self.init['period'] = period

        if rates is not None:
            self.update(rates=rates, period=period, schedule=schedule)

    @property
    def r(self):
        if self.initialized:
            return self._get_cython_attribute("r")
        else:
            Messages._error("Read-out of 'r' is only possible after compile.")

    @r.setter
    def r(self, new_r):
        Messages._error("The value of r is defined through the '*'rates' argument.")

    def update(self, rates, schedule=0., period=-1):
        """
        Set a new list of inputs. The first axis corresponds to time, the others to the desired dimensions of the population. Note, the
        geometry is set during construction phase of the object.

        :param rates: array of firing rates. The first axis corresponds to time, the others to the desired dimensions of the population.
        :param schedule: either a single value or a list of time points where inputs should be set. Default: every timestep.
        :param period: time when the timed array will be reset and start again, allowing cycling over the inputs. Default: no cycling (-1.).
        """
        self.rates = rates
        self.period = period

        # Check the schedule
        if isinstance(schedule, (int, float)):
            if float(schedule) <= 0.0:
                schedule = get_global_config('dt')

            self.schedule = [ float(schedule*i) for i in range(rates.shape[0])]
        else:
            self.schedule = schedule

        if len(self.schedule) > self.rates.shape[0]:
            Messages._error('TimedArray: the length of the schedule parameter cannot exceed the first dimension of the rates parameter.')

        if len(self.schedule) < self.rates.shape[0]:
            Messages._warning('TimedArray: the length of the schedule parameter is smaller than the first dimension of the rates parameter (more data than time points). Make sure it is what you expect.')

    def _copy(self):
        "Returns a copy of the population when creating networks."
        return TimedArray(rates=self.rates, geometry=self.geometry, schedule=self.schedule, period=self.period, name=self.name, copied=True)

    def _generate_st(self):
        """
        adjust code templates for the specific population for single thread and openMP.
        """
        self._specific_template['declare_additional'] = """
    // Custom local parameters of a TimedArray
    std::vector< int > _schedule; // List of times where new inputs should be set
    std::vector< std::vector< %(float_prec)s > > _buffer; // buffer holding the data
    int _period; // Period of cycling
    long int _t; // Internal time
    int _block; // Internal block when inputs are set not at each step
""" % {'float_prec': get_global_config('precision')}
        self._specific_template['access_additional'] = """
    // Custom local parameters of a TimedArray
    void set_schedule(std::vector<int> schedule) { _schedule = schedule; }
    std::vector<int> get_schedule() { return _schedule; }
    void set_buffer(std::vector< std::vector< %(float_prec)s > > buffer) { _buffer = buffer; r = _buffer[_block]; }
    std::vector< std::vector< %(float_prec)s > > get_buffer() { return _buffer; }
    void set_period(int period) { _period = period; }
    int get_period() { return _period; }
""" % {'float_prec': get_global_config('precision')}
        self._specific_template['init_additional'] = """
        // Initialize counters
        _t = 0;
        _block = 0;
        _period = -1;
"""
        self._specific_template['export_additional'] = """
        # Custom local parameters of a TimedArray
        void set_schedule(vector[int])
        vector[int] get_schedule()
        void set_buffer(vector[vector[%(float_prec)s]])
        vector[vector[%(float_prec)s]] get_buffer()
        void set_period(int)
        int get_period()
""" % {'float_prec': get_global_config('precision')}

        self._specific_template['reset_additional'] ="""
        _t = 0;
        _block = 0;

        r.clear();
        r = std::vector<%(float_prec)s>(size, 0.0);
""" % {'float_prec': get_global_config('precision')}

        self._specific_template['wrapper_access_additional'] = """
    # Custom local parameters of a TimedArray
    cpdef set_schedule( self, schedule ):
        pop%(id)s.set_schedule( schedule )
    cpdef np.ndarray get_schedule( self ):
        return np.array(pop%(id)s.get_schedule( ))

    cpdef set_rates( self, buffer ):
        pop%(id)s.set_buffer( buffer )
    cpdef np.ndarray get_rates( self ):
        return np.array(pop%(id)s.get_buffer( ))

    cpdef set_period( self, period ):
        pop%(id)s.set_period(period)
    cpdef int get_period(self):
        return pop%(id)s.get_period()
""" % { 'id': self.id }

        self._specific_template['update_variables'] = """
        if(_active) {
        #ifdef _DEBUG
            std::cout << "TimedArray::update() - " << _t << " " << _block<< " " << _schedule[_block] << std::endl;
        #endif

            // Check if it is time to set the input
            if(_t == _schedule[_block]){
                // Set the data
                r = _buffer[_block];
                // Move to the next block
                _block++;
                // If was the last block, go back to the first block
                if (_block == _schedule.size()){
                    _block = 0;
                }
            }

            // If the timedarray is periodic, check if we arrive at that point
            if(_period > -1 && (_t == _period-1)){
                // Reset the counters
                _block=0;
                _t = -1;
            }

            // Always increment the internal time
            _t++;

        #ifdef _DEBUG
            std::cout << "TimedArray::update(t="<< t <<") - current buffer (min/max) = [" << *std::min_element(r.begin(), r.end()) << "," << *std::max_element(r.begin(), r.end()) <<  "]" << std::endl;
        #endif
        }
"""

        self._specific_template['size_in_bytes'] = """
        // schedule
        size_in_bytes += _schedule.capacity() * sizeof(int);

        // buffer
        size_in_bytes += _buffer.capacity() * sizeof(std::vector<%(float_prec)s>);
        for( auto it = _buffer.begin(); it != _buffer.end(); it++ )
            size_in_bytes += it->capacity() * sizeof(%(float_prec)s);
""" % {'float_prec': get_global_config('precision')}

    def _generate_omp(self):
        """
        adjust code templates for the specific population for single thread and openMP.
        """
        self._specific_template['declare_additional'] = """
    // Custom local parameters of a TimedArray
    std::vector< int > _schedule; // List of times where new inputs should be set
    std::vector< std::vector< %(float_prec)s > > _buffer; // buffer holding the data
    int _period; // Period of cycling
    long int _t; // Internal time
    int _block; // Internal block when inputs are set not at each step
""" % {'float_prec': get_global_config('precision')}
        self._specific_template['access_additional'] = """
    // Custom local parameters of a TimedArray
    void set_schedule(std::vector<int> schedule) { _schedule = schedule; }
    std::vector<int> get_schedule() { return _schedule; }
    void set_buffer(std::vector< std::vector< %(float_prec)s > > buffer) { _buffer = buffer; r = _buffer[_block]; }
    std::vector< std::vector< %(float_prec)s > > get_buffer() { return _buffer; }
    void set_period(int period) { _period = period; }
    int get_period() { return _period; }
""" % {'float_prec': get_global_config('precision')}
        self._specific_template['init_additional'] = """
        // Initialize counters
        _t = 0;
        _block = 0;
        _period = -1;
"""
        self._specific_template['export_additional'] = """
        # Custom local parameters of a TimedArray
        void set_schedule(vector[int])
        vector[int] get_schedule()
        void set_buffer(vector[vector[%(float_prec)s]])
        vector[vector[%(float_prec)s]] get_buffer()
        void set_period(int)
        int get_period()
""" % {'float_prec': get_global_config('precision')}

        self._specific_template['reset_additional'] ="""
        _t = 0;
        _block = 0;

        r.clear();
        r = std::vector<%(float_prec)s>(size, 0.0);
""" % {'float_prec': get_global_config('precision')}

        self._specific_template['wrapper_access_additional'] = """
    # Custom local parameters of a TimedArray
    cpdef set_schedule( self, schedule ):
        pop%(id)s.set_schedule( schedule )
    cpdef np.ndarray get_schedule( self ):
        return np.array(pop%(id)s.get_schedule( ))

    cpdef set_rates( self, buffer ):
        pop%(id)s.set_buffer( buffer )
    cpdef np.ndarray get_rates( self ):
        return np.array(pop%(id)s.get_buffer( ))

    cpdef set_period( self, period ):
        pop%(id)s.set_period(period)
    cpdef int get_period(self):
        return pop%(id)s.get_period()
""" % { 'id': self.id }

        self._specific_template['update_variables'] = """
        if(_active){
            #pragma omp single
            {
            #ifdef _DEBUG
                std::cout << "TimedArray::update() - " << _t << " " << _block<< " " << _schedule[_block] << std::endl;
            #endif

                // Check if it is time to set the input
                if(_t == _schedule[_block]){
                    // Set the data
                    r = _buffer[_block];
                    // Move to the next block
                    _block++;
                    // If was the last block, go back to the first block
                    if (_block == _schedule.size()){
                        _block = 0;
                    }
                }

                // If the timedarray is periodic, check if we arrive at that point
                if(_period > -1 && (_t == _period-1)){
                    // Reset the counters
                    _block=0;
                    _t = -1;
                }

                // Always increment the internal time
                _t++;
            }
        }
"""

        self._specific_template['size_in_bytes'] = """
        // schedule
        size_in_bytes += _schedule.capacity() * sizeof(int);

        // buffer
        size_in_bytes += _buffer.capacity() * sizeof(std::vector<%(float_prec)s>);
        for( auto it = _buffer.begin(); it != _buffer.end(); it++ )
            size_in_bytes += it->capacity() * sizeof(%(float_prec)s);
""" % {'float_prec': get_global_config('precision')}

    def _generate_cuda(self):
        """
        adjust code templates for the specific population for single thread and CUDA.
        """
        # HD (18. Nov 2016)
        # I suppress the code generation for allocating the variable r on gpu, as
        # well as memory transfer codes. This is only possible as no other variables
        # allowed in TimedArray.
        self._specific_template['init_parameters_variables'] = ""
        self._specific_template['host_device_transfer'] = ""
        self._specific_template['device_host_transfer'] = ""

        #
        # Code for handling the buffer and schedule parameters
        self._specific_template['declare_additional'] = """
    // Custom local parameter timed array
    std::vector< int > _schedule;
    std::vector< %(float_prec)s* > gpu_buffer;
    int _period; // Period of cycling
    long int _t; // Internal time
    int _block; // Internal block when inputs are set not at each step
""" % {'float_prec': get_global_config('precision')}
        self._specific_template['access_additional'] = """
    // Custom local parameter timed array
    void set_schedule(std::vector<int> schedule) { _schedule = schedule; }
    std::vector<int> get_schedule() { return _schedule; }
    void set_buffer(std::vector< std::vector< %(float_prec)s > > buffer) {
        if ( gpu_buffer.empty() ) {
            gpu_buffer = std::vector< %(float_prec)s* >(buffer.size(), nullptr);
            // allocate gpu arrays
            for(int i = 0; i < buffer.size(); i++) {
                cudaMalloc((void**)&gpu_buffer[i], buffer[i].size()*sizeof(%(float_prec)s));
            }
        }

        auto host_it = buffer.begin();
        auto dev_it = gpu_buffer.begin();
        for (; host_it < buffer.end(); host_it++, dev_it++) {
            cudaMemcpy( *dev_it, host_it->data(), host_it->size()*sizeof(%(float_prec)s), cudaMemcpyHostToDevice);
        }

        gpu_r = gpu_buffer[_block];
    }
    std::vector< std::vector< %(float_prec)s > > get_buffer() {
        std::vector< std::vector< %(float_prec)s > > buffer = std::vector< std::vector< %(float_prec)s > >( gpu_buffer.size(), std::vector<%(float_prec)s>(size,0.0) );

        auto host_it = buffer.begin();
        auto dev_it = gpu_buffer.begin();
        for (; host_it < buffer.end(); host_it++, dev_it++) {
            cudaMemcpy( host_it->data(), *dev_it, size*sizeof(%(float_prec)s), cudaMemcpyDeviceToHost );
        }

        return buffer;
    }
    void set_period(int period) { _period = period; }
    int get_period() { return _period; }
""" % {'float_prec': get_global_config('precision')}
        self._specific_template['init_additional'] = """
        // counters
        _t = 0;
        _block = 0;
        _period = -1;
"""
        self._specific_template['reset_additional'] = """
        // counters
        _t = 0;
        _block = 0;
        gpu_r = gpu_buffer[0];
"""
        self._specific_template['export_additional'] = """
        # Custom local parameters timed array
        void set_schedule(vector[int])
        vector[int] get_schedule()
        void set_buffer(vector[vector[%(float_prec)s]])
        vector[vector[%(float_prec)s]] get_buffer()
        void set_period(int)
        int get_period()
""" % {'float_prec': get_global_config('precision')}
        self._specific_template['wrapper_access_additional'] = """
    # Custom local parameters timed array
    cpdef set_schedule( self, schedule ):
        pop%(id)s.set_schedule( np.array(schedule, dtype=int) )
    cpdef np.ndarray get_schedule( self ):
        return np.array(pop%(id)s.get_schedule( ))

    cpdef set_rates( self, buffer ):
        pop%(id)s.set_buffer( buffer )
    cpdef np.ndarray get_rates( self ):
        return np.array(pop%(id)s.get_buffer( ))

    cpdef set_period( self, period ):
        pop%(id)s.set_period(period)
    cpdef int get_periodic(self):
        return pop%(id)s.get_period()
""" % { 'id': self.id, 'float_prec': get_global_config('precision') }

        # there is no GPU-side computation
        self._specific_template['update_variable_body'] = ""
        self._specific_template['update_variable_invoke'] = ""
        self._specific_template['update_variable_header'] = ""

        # we switch the GPU buffer which is read out in each time step
        self._specific_template['update_variables'] = """
        if(_active) {
        #ifdef _DEBUG
            std::cout << "TimedArray::update() - " << _t << " " << _block<< " " << _schedule[_block] << std::endl;
        #endif
            // Check if it is time to set the input
            if(_t == _schedule[_block]){
                // Set the data
                gpu_r = gpu_buffer[_block];
                // Move to the next block
                _block++;
                // If was the last block, go back to the first block
                if ( _block == _schedule.size() ) {
                    _block = 0;
                }
            }

            // If the timedarray is periodic, check if we arrive at that point
            if( (_period > -1) && (_t == _period-1) ) {
                // Reset the counters
                _block=0;
                _t = -1;
            }

            // Always increment the internal time
            _t++;
        }
"""
        # call the switch of CPU-buffers (host-side)
        self._specific_template['update_variable_call'] = """
    // host side update of neurons
    pop%(id)s.update();
""" % {'id': self.id}

        self._specific_template['size_in_bytes'] = """
        // r
        size_in_bytes += sizeof(std::vector<%(float_prec)s>);
        size_in_bytes += r.capacity() * sizeof(%(float_prec)s);

        // schedule
        size_in_bytes += sizeof(std::vector<int>);
        size_in_bytes += _schedule.capacity() * sizeof(int);

        // gpu_buffer
        size_in_bytes += sizeof(std::vector<%(float_prec)s*>);
        size_in_bytes += gpu_buffer.capacity() * sizeof(%(float_prec)s*);
""" % {'float_prec': get_global_config('precision')}

    def _instantiate(self, module):
        # Create the Cython instance
        self.cyInstance = getattr(module, self.class_name+'_wrapper')(self.size, self.max_delay)

    def __setattr__(self, name, value):
        if name == 'schedule':
            if self.initialized:
                val_int = np.array((np.atleast_1d(value) / get_global_config('dt')), dtype=np.int32)
                self.cyInstance.set_schedule( val_int )
            else:
                self.init['schedule'] = value
        elif name == 'rates':
            if self.initialized:
                if value is None:
                    return # nothing to do

                if len(value.shape) > 2:
                    if value.shape[1:] != self.geometry:
                        Messages._error("TimedArray: mismatch between *rates* argument (", value.shape[1:], ") and stored geometry (", self.geometry, ").")

                    # we need to flatten the provided data
                    flat_values = value.reshape( (value.shape[0], self.size) )
                    self.cyInstance.set_rates( flat_values )
                else:
                    self.cyInstance.set_rates( value )
            else:
                self.init['rates'] = value
        elif name == "period":
            if self.initialized:
                self.cyInstance.set_period(int(value /get_global_config('dt')))
            else:
                self.init['period'] = value
        else:
            Population.__setattr__(self, name, value)

    def __getattr__(self, name):
        if name == 'schedule':
            if self.initialized:
                return get_global_config('dt') * self.cyInstance.get_schedule()
            else:
                return self.init['schedule']
        elif name == 'rates':
            if self.initialized:
                if len(self.geometry) > 1:
                    # unflatten the data
                    flat_values = self.cyInstance.get_rates()
                    values = np.zeros( tuple( [len(self.schedule)] + list(self.geometry) ) )
                    for x in range(len(self.schedule)):
                        values[x] = np.reshape( flat_values[x], self.geometry)
                    return values
                else:
                    return self.cyInstance.get_rates()
            else:
                return self.init['rates']
        elif name == 'period':
            if self.initialized:
                return self.cyInstance.get_period() * get_global_config('dt')
            else:
                return self.init['period']
        else:
            return Population.__getattribute__(self, name)

class TimedPoissonPopulation(SpecificPopulation):
    """
    Poisson population whose rate vary with the provided schedule.

    Example:

    ```python
    inp = TimedPoissonPopulation(
        geometry = 100,
        rates = [10., 20., 100., 20., 5.],
        schedule = [0., 100., 200., 500., 600.],
    )
    ```


    This creates a population of 100 Poisson neurons whose rate will be:
    
    * 10 Hz during the first 100 ms.
    * 20 HZ during the next 100 ms.
    * 100 Hz during the next 300 ms.
    * 20 Hz during the next 100 ms.
    * 5 Hz until the end of the simulation.
    
    
    If you want the TimedPoissonPopulation to "loop" over the schedule, you can specify a period:

    ```python
    inp = TimedPoissonPopulation(
        geometry = 100,
        rates = [10., 20., 100., 20., 5.],
        schedule = [0., 100., 200., 500., 600.],
        period = 1000.,
    )
    ```

    Here the rate will become 10Hz again every 1 second of simulation. If the period is smaller than the schedule, the remaining rates will not be set.

    Note that you can use the ``reset()`` method to manually reinitialize the schedule, times becoming relative to that call:

    ```python
    simulate(1200.) # Should switch to 100 Hz due to the period of 1000.
    inp.reset()
    simulate(1000.) # Starts at 10 Hz again.
    ```

    The rates were here global to the population. If you want each neuron to have a different rate, ``rates`` must have additional dimensions corresponding to the geometry of the population.

    ```python
    inp = TimedPoissonPopulation(
        geometry = 100,
        rates = [ 
            [10. + 0.05*i for i in range(100)], 
            [20. + 0.05*i for i in range(100)],
        ],
        schedule = [0., 100.],
        period = 1000.,
    )
    ```

    """
    def __init__(self, geometry, rates, schedule, period= -1., name=None, copied=False):
        """    
        :param rates: array of firing rates. The first axis corresponds to the times where the firing rate should change. If a different rate should be used by the different neurons, the other dimensions must match the geometry of the population.
        :param schedule: list of times (in ms) where the firing rate should change.
        :param period: time when the timed array will be reset and start again, allowing cycling over the schedule. Default: no cycling (-1.).
        """
        
        neuron = Neuron(
            parameters = """
            proba = 1.0
            """,
            equations = """
            p = Uniform(0.0, 1.0) * 1000.0 / dt
            """,
            spike = """
            p < proba
            """,
            name="TimedPoisson",
            description="Spiking neuron following a Poisson distribution."
        )

        SpecificPopulation.__init__(self, geometry=geometry, neuron=neuron, name=name, copied=copied)

        # Check arguments
        try:
            rates = np.array(rates)
        except:
            Messages._error("TimedPoissonPopulation: the rates argument must be a numpy array.")

        schedule = np.array(schedule)

        nb_schedules = rates.shape[0]
        if nb_schedules != schedule.size:
            Messages._error("TimedPoissonPopulation: the first axis of the rates argument must be the same length as schedule.")


        if rates.ndim == 1 : # One rate for the whole population
            rates = np.array([np.full(self.size, rates[i]) for i in range(nb_schedules)]) 

        # Initial values
        self.init['schedule'] = schedule
        self.init['rates'] = rates
        self.init['period'] = period

    def _copy(self):
        "Returns a copy of the population when creating networks."
        return TimedPoissonPopulation(self.geometry, self.init['rates'] , self.init['schedule'], self.init['period'], self.name, copied=True)

    def _generate_st(self):
        """
        adjust code templates for the specific population for single thread.
        """
        self._specific_template['declare_additional'] = """
    // Custom local parameters of a TimedPoissonPopulation
    std::vector< int > _schedule; // List of times where new inputs should be set
    std::vector< std::vector< %(float_prec)s > > _buffer; // buffer holding the data
    int _period; // Period of cycling
    long int _t; // Internal time
    int _block; // Internal block when inputs are set not at each step
""" % {'float_prec': get_global_config('precision')}
        self._specific_template['access_additional'] = """
    // Custom local parameters of a TimedPoissonPopulation
    void set_schedule(std::vector<int> schedule) { _schedule = schedule; }
    std::vector<int> get_schedule() { return _schedule; }
    void set_buffer(std::vector< std::vector< %(float_prec)s > > buffer) { _buffer = buffer; r = _buffer[0]; }
    std::vector< std::vector< %(float_prec)s > > get_buffer() { return _buffer; }
    void set_period(int period) { _period = period; }
    int get_period() { return _period; }
""" % {'float_prec': get_global_config('precision')}
        self._specific_template['init_additional'] = """
        // Initialize counters
        _t = 0;
        _block = 0;
        _period = -1;
"""
        self._specific_template['export_additional'] = """
        # Custom local parameters of a TimedPoissonPopulation
        void set_schedule(vector[int])
        vector[int] get_schedule()
        void set_buffer(vector[vector[%(float_prec)s]])
        vector[vector[%(float_prec)s]] get_buffer()
        void set_period(int)
        int get_period()
""" % {'float_prec': get_global_config('precision')}

        self._specific_template['reset_additional'] ="""
        _t = 0;
        _block = 0;

        r.clear();
        r = std::vector<%(float_prec)s>(size, 0.0);
""" % {'float_prec': get_global_config('precision')}

        self._specific_template['wrapper_access_additional'] = """
    # Custom local parameters of a TimedArray
    cpdef set_schedule( self, schedule ):
        pop%(id)s.set_schedule( schedule )
    cpdef np.ndarray get_schedule( self ):
        return np.array(pop%(id)s.get_schedule( ))

    cpdef set_rates( self, buffer ):
        pop%(id)s.set_buffer( buffer )
    cpdef np.ndarray get_rates( self ):
        return np.array(pop%(id)s.get_buffer( ))

    cpdef set_period( self, period ):
        pop%(id)s.set_period(period)
    cpdef int get_period(self):
        return pop%(id)s.get_period()
""" % { 'id': self.id }

        self._specific_template['update_variables'] = """
        if(_active){
            //std::cout << _t << " " << _block<< " " << _schedule[_block] << std::endl;

            // Check if it is time to set the input
            if(_t == _schedule[_block]){
                // Set the data
                proba = _buffer[_block];
                // Move to the next block
                _block++;
                // If was the last block, go back to the first block
                if (_block == _schedule.size()){
                    _block = 0;
                }
            }

            // If the timedarray is periodic, check if we arrive at that point
            if(_period > -1 && (_t == _period-1)){
                // Reset the counters
                _block=0;
                _t = -1;
            }

            // Always increment the internal time
            _t++;
        }

        if( _active ) {
            spiked.clear();

            // Updating local variables
            %(float_prec)s step = 1000.0/dt;

            #pragma omp simd
            for(int i = 0; i < size; i++){

                // p = Uniform(0.0, 1.0) * 1000.0 / dt
                p[i] = step*rand_0[i];


            }
        } // active
""" % {'float_prec': get_global_config('precision')}

        self._specific_template['size_in_bytes'] = """
        // schedule
        size_in_bytes += _schedule.capacity() * sizeof(int);

        // buffer
        size_in_bytes += _buffer.capacity() * sizeof(std::vector<%(float_prec)s>);
        for( auto it = _buffer.begin(); it != _buffer.end(); it++ )
            size_in_bytes += it->capacity() * sizeof(%(float_prec)s);
""" % {'float_prec': get_global_config('precision')}

    def _generate_omp(self):
        """
        adjust code templates for the specific population for openMP.
        """
        self._specific_template['declare_additional'] = """
    // Custom local parameters of a TimedPoissonPopulation
    std::vector< int > _schedule; // List of times where new inputs should be set
    std::vector< std::vector< %(float_prec)s > > _buffer; // buffer holding the data
    int _period; // Period of cycling
    long int _t; // Internal time
    int _block; // Internal block when inputs are set not at each step
""" % {'float_prec': get_global_config('precision')}
        self._specific_template['access_additional'] = """
    // Custom local parameters of a TimedPoissonPopulation
    void set_schedule(std::vector<int> schedule) { _schedule = schedule; }
    std::vector<int> get_schedule() { return _schedule; }
    void set_buffer(std::vector< std::vector< %(float_prec)s > > buffer) { _buffer = buffer; r = _buffer[0]; }
    std::vector< std::vector< %(float_prec)s > > get_buffer() { return _buffer; }
    void set_period(int period) { _period = period; }
    int get_period() { return _period; }
""" % {'float_prec': get_global_config('precision')}
        self._specific_template['init_additional'] = """
        // Initialize counters
        _t = 0;
        _block = 0;
        _period = -1;
"""
        self._specific_template['export_additional'] = """
        # Custom local parameters of a TimedPoissonPopulation
        void set_schedule(vector[int])
        vector[int] get_schedule()
        void set_buffer(vector[vector[%(float_prec)s]])
        vector[vector[%(float_prec)s]] get_buffer()
        void set_period(int)
        int get_period()
""" % {'float_prec': get_global_config('precision')}

        self._specific_template['reset_additional'] ="""
        _t = 0;
        _block = 0;

        r.clear();
        r = std::vector<%(float_prec)s>(size, 0.0);
""" % {'float_prec': get_global_config('precision')}

        self._specific_template['wrapper_access_additional'] = """
    # Custom local parameters of a TimedArray
    cpdef set_schedule( self, schedule ):
        pop%(id)s.set_schedule( schedule )
    cpdef np.ndarray get_schedule( self ):
        return np.array(pop%(id)s.get_schedule( ))

    cpdef set_rates( self, buffer ):
        pop%(id)s.set_buffer( buffer )
    cpdef np.ndarray get_rates( self ):
        return np.array(pop%(id)s.get_buffer( ))

    cpdef set_period( self, period ):
        pop%(id)s.set_period(period)
    cpdef int get_period(self):
        return pop%(id)s.get_period()
""" % { 'id': self.id }

        self._specific_template['update_variables'] = """
        if(_active){
            #pragma omp single
            {
                //std::cout << _t << " " << _block<< " " << _schedule[_block] << std::endl;

                // Check if it is time to set the input
                if(_t == _schedule[_block]){
                    // Set the data
                    proba = _buffer[_block];
                    // Move to the next block
                    _block++;
                    // If was the last block, go back to the first block
                    if (_block == _schedule.size()){
                        _block = 0;
                    }
                }

                // If the timedarray is periodic, check if we arrive at that point
                if(_period > -1 && (_t == _period-1)){
                    // Reset the counters
                    _block=0;
                    _t = -1;
                }

                // Always increment the internal time
                _t++;
            }
        }

        if( _active ) {
            spiked.clear();

            // Updating local variables
            %(float_prec)s step = 1000.0/dt;

            #pragma omp for simd
            for(int i = 0; i < size; i++){

                // p = Uniform(0.0, 1.0) * 1000.0 / dt
                p[i] = step*rand_0[i];


            }
        } // active
""" % {'float_prec': get_global_config('precision')}

        self._specific_template['size_in_bytes'] = """
        // schedule
        size_in_bytes += _schedule.capacity() * sizeof(int);

        // buffer
        size_in_bytes += _buffer.capacity() * sizeof(std::vector<%(float_prec)s>);
        for( auto it = _buffer.begin(); it != _buffer.end(); it++ )
            size_in_bytes += it->capacity() * sizeof(%(float_prec)s);
""" % {'float_prec': get_global_config('precision')}

    def _generate_cuda(self):
        """
        Code generation if the CUDA paradigm is set.
        """
        # I suppress the code generation for allocating the variable r on gpu, as
        # well as memory transfer codes. This is only possible as no other variables
        # allowed in TimedArray.
        self._specific_template['init_parameters_variables'] = """
        // Random numbers
        cudaMalloc((void**)&gpu_rand_0, size * sizeof(curandState));
        init_curand_states( size, gpu_rand_0, global_seed );
"""
        self._specific_template['host_device_transfer'] = ""
        self._specific_template['device_host_transfer'] = ""

        #
        # Code for handling the buffer and schedule parameters
        self._specific_template['declare_additional'] = """
    // Custom local parameter timed array
    std::vector< int > _schedule;
    std::vector< %(float_prec)s* > gpu_buffer;
    int _period; // Period of cycling
    long int _t; // Internal time
    int _block; // Internal block when inputs are set not at each step
""" % {'float_prec': get_global_config('precision')}
        self._specific_template['access_additional'] = """
    // Custom local parameter timed array
    void set_schedule(std::vector<int> schedule) { _schedule = schedule; }
    std::vector<int> get_schedule() { return _schedule; }
    void set_buffer(std::vector< std::vector< %(float_prec)s > > buffer) {
        if ( gpu_buffer.empty() ) {
            gpu_buffer = std::vector< %(float_prec)s* >(buffer.size(), nullptr);
            // allocate gpu arrays
            for(int i = 0; i < buffer.size(); i++) {
                cudaMalloc((void**)&gpu_buffer[i], buffer[i].size()*sizeof(%(float_prec)s));
            }
        }

        auto host_it = buffer.begin();
        auto dev_it = gpu_buffer.begin();
        for( ; host_it < buffer.end(); host_it++, dev_it++ ) {
            cudaMemcpy( *dev_it, host_it->data(), host_it->size()*sizeof(%(float_prec)s), cudaMemcpyHostToDevice);
        }

        gpu_proba = gpu_buffer[0];
    }
    std::vector< std::vector< %(float_prec)s > > get_buffer() {
        std::vector< std::vector< %(float_prec)s > > buffer = std::vector< std::vector< %(float_prec)s > >( gpu_buffer.size(), std::vector<%(float_prec)s>(size,0.0) );

        auto host_it = buffer.begin();
        auto dev_it = gpu_buffer.begin();
        for( ; host_it < buffer.end(); host_it++, dev_it++ ) {
            cudaMemcpy( host_it->data(), *dev_it, size*sizeof(%(float_prec)s), cudaMemcpyDeviceToHost );
        }

        return buffer;
    }
    void set_period(int period) { _period = period; }
    int get_period() { return _period; }
""" % {'float_prec': get_global_config('precision')}
        self._specific_template['init_additional'] = """
        // counters
        _t = 0;
        _block = 0;
        _period = -1;
"""
        self._specific_template['reset_additional'] = """
        // counters
        _t = 0;
        _block = 0;
        gpu_proba = gpu_buffer[0];
"""
        self._specific_template['export_additional'] = """
        # Custom local parameters timed array
        void set_schedule(vector[int])
        vector[int] get_schedule()
        void set_buffer(vector[vector[%(float_prec)s]])
        vector[vector[%(float_prec)s]] get_buffer()
        void set_period(int)
        int get_period()
""" % {'float_prec': get_global_config('precision')}
        self._specific_template['wrapper_access_additional'] = """
    # Custom local parameters timed array
    cpdef set_schedule( self, schedule ):
        pop%(id)s.set_schedule( schedule )
    cpdef np.ndarray get_schedule( self ):
        return np.array(pop%(id)s.get_schedule( ))

    cpdef set_rates( self, buffer ):
        pop%(id)s.set_buffer( buffer )
    cpdef np.ndarray get_rates( self ):
        return np.array(pop%(id)s.get_buffer( ))

    cpdef set_period( self, period ):
        pop%(id)s.set_period(period)
    cpdef int get_periodic(self):
        return pop%(id)s.get_period()
""" % { 'id': self.id, 'float_prec': get_global_config('precision') }

        self._specific_template['update_variables'] = """
        if(_active) {
            // std::cout << _t << " " << _block<< " " << _schedule[_block] << std::endl;
            // Check if it is time to set the input
            if(_t == _schedule[_block]){
                // Set the data
                gpu_proba = gpu_buffer[_block];
                // Move to the next block
                _block++;
                // If was the last block, go back to the first block
                if ( _block == _schedule.size() ) {
                    _block = 0;
                }
            }

            // If the timedarray is periodic, check if we arrive at that point
            if( (_period > -1) && (_t == _period-1) ) {
                // Reset the counters
                _block=0;
                _t = -1;
            }

            // Always increment the internal time
            _t++;
        }
"""

        self._specific_template['update_variable_body'] = """
__global__ void cuPop%(id)s_local_step( const long int t, const double dt, curandState* rand_0, double* proba, unsigned int* num_events, int* spiked, long int* last_spike )
{
    int i = threadIdx.x + blockDim.x * blockIdx.x;
    %(float_prec)s step = 1000.0/dt;

    while ( i < %(size)s )
    {
        // p = Uniform(0.0, 1.0) * 1000.0 / dt
        %(float_prec)s p = curand_uniform_double( &rand_0[i] ) * step;

        if (p < proba[i]) {
            int pos = atomicAdd ( num_events, 1);
            spiked[pos] = i;
            last_spike[i] = t;
        }

        i += blockDim.x;
    }

    __syncthreads();
}
""" % {
    'id': self.id,
    'size': self.size,
    'float_prec': get_global_config('precision')
}

        self._specific_template['update_variable_header'] = "__global__ void cuPop%(id)s_local_step( const long int t, const double dt, curandState* rand_0, double* proba, unsigned int* num_events, int* spiked, long int* last_spike );" % {'id': self.id}
        # Please notice, that the GPU kernels can be launched only with one block. Otherwise, the
        # atomicAdd which is called inside the kernel is not working correct (HD: April 1st, 2021)
        self._specific_template['update_variable_call'] = """
    // host side update of neurons
    pop%(id)s.update();

    // Reset old events
    clear_num_events<<< 1, 1, 0, pop%(id)s.stream >>>(pop%(id)s.gpu_spike_count);
#ifdef _DEBUG
    cudaError_t err_clear_num_events_%(id)s = cudaGetLastError();
    if(err_clear_num_events_%(id)s != cudaSuccess)
        std::cout << "pop%(id)s_spike_gather: " << cudaGetErrorString(err_clear_num_events_%(id)s) << std::endl;
#endif

    // Compute current events
    cuPop%(id)s_local_step<<< 1, pop%(id)s._threads_per_block, 0, pop%(id)s.stream >>>(
        t, dt,
        pop%(id)s.gpu_rand_0,
        pop%(id)s.gpu_proba,
        pop%(id)s.gpu_spike_count,
        pop%(id)s.gpu_spiked,
        pop%(id)s.gpu_last_spike
    );
#ifdef _DEBUG
    cudaError_t err_pop_spike_gather_%(id)s = cudaGetLastError();
    if(err_pop_spike_gather_%(id)s != cudaSuccess)
        std::cout << "pop%(id)s_spike_gather: " << cudaGetErrorString(err_pop_spike_gather_%(id)s) << std::endl;
#endif

    // transfer back the spike counter (needed by record)
    cudaMemcpyAsync( &pop%(id)s.spike_count, pop%(id)s.gpu_spike_count, sizeof(unsigned int), cudaMemcpyDeviceToHost, pop%(id)s.stream );
#ifdef _DEBUG
    cudaError_t err = cudaGetLastError();
    if ( err != cudaSuccess )
        std::cout << "record_spike_count: " << cudaGetErrorString(err) << std::endl;
#endif

    // transfer back the spiked array (needed by record)
    cudaMemcpyAsync( pop%(id)s.spiked.data(), pop%(id)s.gpu_spiked, pop%(id)s.spike_count*sizeof(int), cudaMemcpyDeviceToHost, pop%(id)s.stream );
#ifdef _DEBUG
    err = cudaGetLastError();
    if ( err != cudaSuccess )
        std::cout << "record_spike: " << cudaGetErrorString(err) << std::endl;
#endif

""" % {'id': self.id}

        self._specific_template['size_in_bytes'] = "//TODO: "

    def _instantiate(self, module):
        # Create the Cython instance
        self.cyInstance = getattr(module, self.class_name+'_wrapper')(self.size, self.max_delay)

    def __setattr__(self, name, value):
        if name == 'schedule':
            if self.initialized:
                self.cyInstance.set_schedule( value / get_global_config('dt') )
            else:
                self.init['schedule'] = value
        elif name == 'rates':
            if self.initialized:
                value = np.array(value)
                if value.shape[0] != self.schedule.shape[0]:
                    Messages._error("TimedPoissonPopulation: the first dimension of rates must match the schedule.")
                if value.ndim > 2:
                    # we need to flatten the provided data
                    values = value.reshape( (value.shape[0], self.size) )
                    self.cyInstance.set_rates(values)
                elif value.ndim == 2:
                    if value.shape[1] != self.size:
                        if value.shape[1] == 1:
                            value = np.array([np.full(self.size, value[i]) for i in range(value.shape[0])])
                        else:
                            Messages._error("TimedPoissonPopulation: the second dimension of rates must match the number of neurons.")
                    self.cyInstance.set_rates(value)
                elif value.ndim == 1:
                    value = np.array([np.full(self.size, value[i]) for i in range(value.shape[0])]) 
                    self.cyInstance.set_rates(value)

            else:
                self.init['rates'] = value
        elif name == "period":
            if self.initialized:
                self.cyInstance.set_period(int(value /get_global_config('dt')))
            else:
                self.init['period'] = value
        else:
            Population.__setattr__(self, name, value)

    def __getattr__(self, name):
        if name == 'schedule':
            if self.initialized:
                return get_global_config('dt') * self.cyInstance.get_schedule()
            else:
                return self.init['schedule']
        elif name == 'rates':
            if self.initialized:
                if len(self.geometry) > 1:
                    # unflatten the data
                    flat_values = self.cyInstance.get_rates()
                    values = np.zeros( tuple( [len(self.schedule)] + list(self.geometry) ) )
                    for x in range(len(self.schedule)):
                        values[x] = np.reshape( flat_values[x], self.geometry)
                    return values
                else:
                    return self.cyInstance.get_rates()
            else:
                return self.init['rates']
        elif name == 'period':
            if self.initialized:
                return self.cyInstance.get_period() * get_global_config('dt')
            else:
                return self.init['period']
        else:
            return Population.__getattribute__(self, name)


"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from ANNarchy.intern import ConfigManagement
from ANNarchy.intern import Messages

import os
import shutil

class NetworkManager :
    """
    This class implements the management of the data for the different networks. The
    add/remove methods allow the 'random' removal of the network instances and later
    refill of empty spaces.

    The class is implemented as singleton and therefore initialized on first request.

    Individual network are accessed by their id and allow the access to their components.
    """
    _instance = None    # singleton instance

    def __init__(self):
        """
        Constructor.
        """
        pass

    def __new__(cls):
        """
        First call construction of the NetworkManager. No additional arguments are required.
        """
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._create_initial_state(cls._instance)
        
        return cls._instance

    def _create_initial_state(self):
        """
        Initialize the container for the initial network.

        Called either from __init__ or clear(). The first
        slot is reserved for the magic network.
        """
        self._network_desc = [
            {
                'populations': [],
                'projections': [],
                'monitors': [],
                'extensions': [],
                'instance': None,
                'compiled': False,
                'directory': None
            },
        ]
        self._py_instances = [None]

    def get_network_dict(self, net_id):
        if net_id < len(self._network_desc):
            return self._network_desc[net_id]
        else:
            Messages._error("Network", net_id, "not existing ...")

    def _get_network_ids(self):
        res = []
        for net in self._py_instances:
            if net is not None:
                res.append(net.id)
        return res

    def _remove_last_item_from_list(self, net_id, list_name):
        if net_id >= len(self._network_desc):
            Messages._error("Network", net_id, "not existing ...")

        if list_name not in self._network_desc[net_id].keys():
            Messages._error("Field", list_name, "not existing ...")

        self._network_desc[net_id][list_name].pop(-1)

    def __len__(self):
        """
        Called if len() is applied on a NetworkManager class.
        """
        return len(self._network_desc)

    def __repr__(self):
        """
        Instead of showing the object pointer, we present the information of
        registered networks.
        """
        string = "<{}.{} object at {}>\n".format( self.__class__.__module__, self.__class__.__name__, hex(id(self)))

        string += "Number of registered networks = " + str(len(self)) + "\n"

        for net_id in range(len(self._network_desc)):
            string += "Network " + str(net_id) + (" (MagicNetwork)" if net_id == 0 else " ") + "\n"
            string += "  populations = ["
            for pop in self._network_desc[net_id]['populations']:
                string += pop.__class__.__name__ + " at " + hex(id(pop)) + ", "
            string += "]\n"

            string += "  projections = ["
            for proj in self._network_desc[net_id]['projections']:
                string += proj.__class__.__name__ + " at " + hex(id(proj)) + ", "
            string += "]\n"

            string += "  monitors = ["
            for mon in self._network_desc[net_id]['monitors']:
                string += mon.__class__.__name__ + " at " + hex(id(mon)) + ", "
            string += "]\n"

            string += "  extensions = ["
            for mon in self._network_desc[net_id]['extensions']:
                string += mon.__class__.__name__ + " at " + hex(id(mon)) + ", "
            string += "]\n"

            string += "  cyInstance = " + str(self._network_desc[net_id]['instance']) + "\n"

        return string

    def add_network(self, py_instance):
        """
        Adds an empty structure for a new network and returns the new network ID.
        """
        new_dict = {
            'populations': [],
            'projections': [],
            'monitors': [],
            'extensions': [],
            'instance': None,
            'compiled': False,
            'directory': None
        }

        found = -1
        # scan for slots which were freed before
        for i, entry in enumerate(self._network_desc):
            if entry == None and self._py_instances[i] == None:
                found = i
                break

        # dependent on the scan append at the end
        # or fill free slot
        if found == -1:
            new_id = len(self._network_desc)
            self._network_desc.append(new_dict)
            self._py_instances.append(py_instance)
        else:
            new_id = found
            self._network_desc[new_id] = new_dict
            self._py_instances[new_id] = py_instance

        Messages._debug("Added network", new_id)
        return new_id

    def _remove_network(self, py_instance):
        """
        Remove the given network from the list of compilable/instantiable networks.
        It is important to invalidate only the slot. If del is called on the dictionary
        entry then this will lead to a removal of the space and therefore all succesequent
        networks would be assigned wrong.

        This function will be called from the Network.__del__() method, after destruction
        of the attached objects.
        """
        Messages._debug("Remove network", py_instance)

        net_id = -1
        for net_id, inst in enumerate(self._py_instances):
            if inst == py_instance:
                self._network_desc[net_id] = None
                self._py_instances[net_id] = None

    def clear(self):
        """
        Remove all assigned networks and restore the initial state.
        """
        # destroy the magic network. The other networks are
        # destroyed through the Network.__del__()
        for pop in self._network_desc[0]['populations']:
            pop._clear()

        for proj in self._network_desc[0]['projections']:
            proj._clear()

        for mon in self._network_desc[0]['monitors']:
            mon._clear()

        # In some cases, we dont want to remove
        disable_rm_directory = ConfigManagement.get_global_config('debug') or ConfigManagement.get_global_config('disable_shared_library_time_offset')
        if disable_rm_directory:
            pass

        elif self._network_desc[0]['directory'] != None:
            # Removes the library used in last running instance
            if os.path.isfile(self._network_desc[0]['directory']+'/ANNarchyCore' + str(0) + '.so'):
                os.remove(self._network_desc[0]['directory']+'/ANNarchyCore' + str(0) + '.so')

            try:
                if os.path.isdir(self._network_desc[0]['directory']):
                    os.rmdir(self._network_desc[0]['directory'])

            except OSError as err:
                # we notice a not empty directory error
                if err.errno == 39:
                    if ConfigManagement.get_global_config('debug') or ConfigManagement.get_global_config('verbose'):
                        Messages._warning("Attempted to clear:", self._network_desc[0]['directory'], "using os.rmdir failed ... retry with shutil")

                    # we re-try it with shutil, if it again fails, we ignore it ...
                    shutil.rmtree(self._network_desc[0]['directory'], ignore_errors=True)

                else:
                    # Re-throw other errors
                    raise

            self._network_desc[0]['directory'] = None

        # This will trigger as last consequence
        # Network.__del__()
        del self._network_desc
        self._create_initial_state()

    ################################
    ## Memory management
    ################################
    def _cpp_memory_footprint(self, net_id):
        """
        Print the C++ memory consumption for populations, projections on the console.

        :param net_id: net_id of the requested network.
        """
        from ANNarchy.core.Global import _bytes_human_readable

        print("Memory consumption of C++ objects (Network {id}): ".format(id=net_id))

        for pop in self.get_populations(net_id=net_id):
            print(pop.name, _bytes_human_readable(pop.size_in_bytes()))

        for proj in self.get_projections(net_id=net_id):
            print(proj.name, _bytes_human_readable(proj.size_in_bytes()))

        for mon in NetworkManager().get_monitors(net_id=net_id):
            print(mon.name, _bytes_human_readable(mon.size_in_bytes()))

    ################################
    ## Population objects
    ################################
    def get_population(self, net_id, name):
        if net_id < len(self._network_desc):
            for pop in self._network_desc[net_id]['populations']:
                if pop.name == name:
                    return pop
            return None
        else:
            Messages._error("Network", net_id, "not existing ...")

    def get_populations(self, net_id):
        if net_id < len(self._network_desc):
            return self._network_desc[net_id]['populations']
        else:
            Messages._error("Network", net_id, "not existing ...")

    def add_population(self, net_id, population):
        if net_id < len(self._network_desc):
            self._network_desc[net_id]['populations'].append(population)
        else:
            Messages._error("Network", net_id, "not existing ...")

    def number_populations(self, net_id):
        if net_id < len(self._network_desc):
            return len(self._network_desc[net_id]['populations'])
        else:
            Messages._error("Network", net_id, "not existing ...")

    ################################
    ## Projection objects
    ################################
    def get_projections(self, net_id, pre=None, post=None, target=None, suppress_error=False) -> list:
        """
        Return the projections attached to network *net_id*. The returned list can be restricted by the
        arguments *pre*, *post*, or *target*.

        HINT: the *suppress_error* flag should only set to *True* in seldom cases.
        """
        if net_id < len(self._network_desc):
            # None of the arguments is set, so we return all
            if post is None and pre is None and target is None:
                return self._network_desc[net_id]['projections']

            # We need to collect the projections according to the criteria
            res = []

            # The user can provide an object or the name, however, the following code
            # expects the population objects.
            if isinstance(post, str):
                post = self.get_population(post, net_id)
            if isinstance(pre, str):
                pre = self.get_population(pre, net_id)
            # Sanity check
            if post is None or pre is None:
                Messages._error("Either post- or pre-synaptic population was not found")

            # All criterias are used
            if post is not None and pre is not None and target is not None:
                for proj in self._network_desc[net_id]['projections']:
                    if proj.post == post and proj.pre == pre and proj.target == target:
                        res.append(proj)

            # post is the criteria
            elif (post is not None) and (pre is None) and (target is None) :
                for proj in self._network_desc[net_id]['projections']:
                    if proj.post == post:
                        res.append(proj)

            # pre is the criteria
            elif (pre is not None) and (post is None) and (target is None):
                for proj in self._network_desc[net_id]['projections']:
                    if proj.pre == pre:
                        res.append(proj)

            # target is the criteria
            elif (target is not None) and (post is None) and (pre is None):
                for proj in self._network_desc[net_id]['projections']:
                    if proj.target == target:
                        res.append(proj)

            # pre/target is the criteria
            elif (pre is not None) and (target is not None) and (post is None) :
                for proj in self._network_desc[net_id]['projections']:
                    if proj.pre == pre and proj.target == target:
                        res.append(proj)

            # post/target is the criteria
            elif (post is not None) and (target is not None) and (pre is None):
                for proj in self._network_desc[net_id]['projections']:
                    if proj.post == post and proj.target == target:
                        res.append(proj)

            # post/pre is the criteria
            elif (post is not None) and (pre is not None) and (target is None):
                for proj in self._network_desc[net_id]['projections']:
                    if proj.post == post and proj.pre == pre:
                        res.append(proj)

            else:
                # for sanity reasons, should not be reached
                raise NotImplementedError

            if not suppress_error and len(res)==0:
                Messages._error("Could not find projections fitting post={post}, pre={pre}, and target={target}"%{'post':post, 'pre': post, 'target':target})

            return res

        else:
            # The network was not in list, we either throw an Exception or return an empty list.
            if not suppress_error:
                Messages._error("Network", net_id, "not existing ...")
            else:
                return []

    def add_projection(self, net_id, projection):
        if net_id < len(self._network_desc):
            self._network_desc[net_id]['projections'].append(projection)
        else:
            Messages._error("Network", net_id, "not existing ...")

    def number_projections(self, net_id):
        if net_id < len(self._network_desc):
            return len(self._network_desc[net_id]['projections'])
        else:
            Messages._error("Network", net_id, "not existing ...")

    ################################
    ## Monitor objects
    ################################
    def get_monitors(self, net_id):
        if net_id < len(self._network_desc):
            return self._network_desc[net_id]['monitors']
        else:
            Messages._error("Network", net_id, "not existing ...")

    def add_monitor(self, net_id, monitor):
        if net_id < len(self._network_desc):
            self._network_desc[net_id]['monitors'].append(monitor)
        else:
            Messages._error("Network", net_id, "not existing ...")

    def number_monitors(self, net_id):
        if net_id < len(self._network_desc):
            return len(self._network_desc[net_id]['monitors'])
        else:
            Messages._error("Network", net_id, "not existing ...")

    ################################
    ## Extensions
    ################################
    def get_extensions(self, net_id):
        if net_id < len(self._network_desc):
            return self._network_desc[net_id]['extensions']
        else:
            Messages._error("Network", net_id, "not existing ...")

    def add_extension(self, net_id, extension):
        if net_id < len(self._network_desc):
            self._network_desc[net_id]['extensions'].append(extension)
        else:
            Messages._error("Network", net_id, "not existing ...")

    def number_extensions(self, net_id):
        if net_id < len(self._network_desc):
            return len(self._network_desc[net_id]['extensions'])
        else:
            Messages._error("Network", net_id, "not existing ...")

    ################################
    ## Code generation
    ################################
    def is_compiled(self, net_id):
        if net_id < len(self._network_desc):
            return self._network_desc[net_id]['compiled']
        else:
            Messages._error("Network", net_id, "not existing ...")

    def set_compiled(self, net_id):
        if net_id < len(self._network_desc):
            self._network_desc[net_id]['compiled'] = True
        else:
            Messages._error("Network", net_id, "not existing ...")

    def cy_instance(self, net_id):
        if net_id < len(self._network_desc):
            return self._network_desc[net_id]['instance']
        else:
            Messages._error("Network", net_id, "not existing ...")

    def get_code_directory(self, net_id):
        if net_id < len(self._network_desc):
            return self._network_desc[net_id]['directory']
        else:
            Messages._error("Network", net_id, "not existing ...")

    def set_code_directory(self, net_id, directory):
        if net_id < len(self._network_desc):
            self._network_desc[net_id]['directory'] = directory
        else:
            Messages._error("Network", net_id, "not existing ...")

    def set_cy_instance(self, net_id, instance):
        if net_id < len(self._network_desc):
            self._network_desc[net_id]['instance'] = instance
        else:
            Messages._error("Network", net_id, "not existing ...")



# Generic imports
import os, sys
import numpy as np

# ANNarchy core
from .core.Global import *
from .core.Simulate import *
from .core.Constant import Constant
from .core.Neuron import Neuron
from .core.Synapse import Synapse
from .core.Population import Population
from .core.Projection import Projection
from .inputs import *
from .core.Dendrite import Dendrite
from .core.Random import Uniform, DiscreteUniform, Normal, LogNormal, Gamma, Exponential, Binomial
from .core.IO import save, load, load_parameter, load_parameters, save_parameters
from .core.Utils import sparse_random_matrix, sparse_delays_from_weights
from .core.Monitor import *
from .core.Network import Network, parallel_run
from .parser.report.Report import report
from .models.Neurons import *
from .models.Synapses import *
from .extensions import *
from .intern.ConfigManagement import setup

# Cython modules
try:
    # HD: until version 4.6 the connectivity class wasn't named properly. To ensure backward compability
    #     we rename the LILConnectivity to CSR
    from .cython_ext import LILConnectivity
    from .cython_ext import LILConnectivity as CSR
except Exception as e:
    print(e)
    print("""
Warning: Cython modules can not be imported. If you are installing ANNarchy, this is normal, ignore this message. If ANNarchy is already installed, something went wrong with the compilation, try reinstalling.
""")

# ANNarchy compilation
from .generator import compile

# several setup() arguments can be set on command-line
from ANNarchy.generator.CmdLineArgParser import CmdLineArgParser
_arg_parser = CmdLineArgParser()
_arg_parser.parse_arguments_for_setup()

# Automatically call ANNarchy.core.Global.clear()
# if the script terminates
import atexit
atexit.register(check_profile_results)
atexit.register(clear)

# Version
__version__ = '4.8'
__release__ = '4.8.1'

print( 'ANNarchy ' + __version__ + ' (' + __release__ + \
                    ') on ' + sys.platform + ' (' + os.name + ').' )



import ANNarchy.parser.report.LatexParser as LatexParser
from ANNarchy.intern.NetworkManager import NetworkManager
from ANNarchy.intern.GlobalObjects import GlobalObjectManager
from ANNarchy.intern import Messages
from ANNarchy.extensions.bold.BoldModel import BoldModel

from ANNarchy.core.Synapse import Synapse
from ANNarchy.core.Neuron import Neuron

import numpy as np
import os

##################################
### Templates
##################################

header = r"""
%  LaTeX file for generating the Model Description Table in Fig. 5 of
%
%  Nordlie E, Gewaltig M-O, Plesser HE (2009)
%  Towards Reproducible Descriptions of Neuronal Network Models.
%  PLoS Comput Biol 5(8): e1000456.
%
%  Paper URL : http://dx.doi.org/10.1371/journal.pcbi.1000456
%  Figure URL: http://dx.doi.org/10.1371/journal.pcbi.1000456.g005
%
%  This file is released under a
%
%   Creative Commons Attribution, non-commercial, share-alike licence
%   http://creativecommons.org/licenses/by-nc-sa/3.0/de/deed.en
%
%  with the following specifications:
%
%  1. When publishing tables generated from this LaTeX file and modified
%     versions of it, you must cite the paper by Nordlie et al given above.
%
%  2. The non-commercial clause applies only to the distribution of THIS FILE
%     and LaTeX source code files derived from it. You may commercially publish
%     documents generated using this file and derivatived versions of this file.
%
%  Contact: Hans Ekkehard Plesser, UMB (hans.ekkehard.plesser at umb.no)
"""

preamble = r"""
\documentclass{article}
\usepackage[margin=1in]{geometry}
\usepackage{tabularx}
\usepackage{multirow}
\usepackage{colortbl}

\usepackage[fleqn]{amsmath}
\setlength{\mathindent}{0em}
%%\usepackage{mathpazo}
\usepackage{breqn}

\usepackage[scaled=.95]{helvet}
\renewcommand\familydefault{\sfdefault}

\renewcommand\arraystretch{1.2}
\pagestyle{empty}

\newcommand{\hdr}[3]{
    \multicolumn{#1}{|l|}{
        \color{white}\cellcolor[gray]{0.0}
        \textbf{\makebox[0pt]{#2}\hspace{0.5\linewidth}\makebox[0pt][c]{#3}}
    }
}

\begin{document}
"""

summary_template = r"""
\noindent
\begin{tabularx}{\linewidth}{|l|X|}\hline
\hdr{2}{A}{Model Summary}\\ \hline
\textbf{Populations}     & %(population_names)s \\ \hline
\textbf{Topology}        & --- \\ \hline
\textbf{Connectivity}    & %(connectivity)s \\ \hline
\textbf{Neuron models}   & %(neuron_models)s \\ \hline
\textbf{Channel models}  & --- \\ \hline
\textbf{Synapse models}  & --- \\ \hline
\textbf{Plasticity}      & %(synapse_models)s\\ \hline
\textbf{Input}           & --- \\ \hline
\textbf{Measurements}    & %(measurements)s \\ \hline
\end{tabularx}

\vspace{2ex}
"""

populations_template = r"""
\noindent
\begin{tabularx}{\linewidth}{|l|l|X|}\hline
\hdr{3}{B}{Populations}\\ \hline
    \textbf{Name}   & \textbf{Elements} & \textbf{Size} \\ \hline
%(populations_description)s
\end{tabularx}

\vspace{2ex}
"""

connectivity_template = r"""
\noindent
\begin{tabularx}{\linewidth}{|l|l|l|X|X|}\hline
\hdr{5}{C}{Connectivity}\\ \hline
\textbf{Source} & \textbf{Destination} & \textbf{Target} & \textbf{Synapse} & \textbf{Pattern} \\ \hline
%(projections_description)s
\end{tabularx}

\vspace{2ex}
"""


parameters_template = r"""
\noindent
\begin{tabularx}{\linewidth}{|X|}\hline
\hdr{1}{F}{Parameters}\\ \hline
\\ \hline
\end{tabularx}
\vspace{2ex}
"""

constants_template = r"""
\noindent
\begin{tabularx}{\linewidth}{|p{0.25\linewidth}|p{0.25\linewidth}|X|}\hline
\textbf{Constants} &\textbf{Name} & \textbf{Value}   \\ \hline
%(parameters)s
\end{tabularx}

\vspace{2ex}
"""

functions_template = r"""
\noindent
\begin{tabularx}{\linewidth}{|p{0.25\linewidth}|X|}\hline
\textbf{Functions} &  
%(parameters)s
\\ \hline
\end{tabularx}

\vspace{2ex}
"""

popparameters_template = r"""
\noindent
\begin{tabularx}{\linewidth}{|p{0.25\linewidth}|p{0.25\linewidth}|X|}\hline
\textbf{Population} & \textbf{Parameter} & \textbf{Value}   \\ \hline
%(parameters)s
\end{tabularx}

\vspace{2ex}
"""

projparameters_template = r"""
\noindent
\begin{tabularx}{\linewidth}{|p{0.25\linewidth}|p{0.25\linewidth}|X|}\hline
\textbf{Projection} & \textbf{Parameter} & \textbf{Value}   \\ \hline
%(parameters)s
\end{tabularx}

\vspace{2ex}
"""

input_template = r"""
\noindent
\begin{tabularx}{\linewidth}{|l|X|}\hline
\hdr{2}{G}{Input}\\ \hline
\textbf{Type} & \textbf{Description} \\ \hline
---
\\ \hline
\end{tabularx}

\vspace{2ex}

"""

footer = r"""
\end{document}
"""

##################################
### Main method
##################################

def report_latex(filename="./report.tex", standalone=True, gather_subprojections=False, net_id=0):
    """   
    Generates a .tex file describing the network according to:

    > Nordlie E, Gewaltig M-O, Plesser HE (2009). Towards Reproducible Descriptions of Neuronal Network Models. PLoS Comput Biol 5(8): e1000456.

    :param filename: name of the .tex file where the report will be written (default: "./report.tex")
    :param standalone: tells if the generated file should be directly compilable or only includable (default: True)
    :param gather_subprojections: if a projection between two populations has been implemented as a multiple of projections between sub-populations, this flag allows to group them in the summary (default: False).
    """

    # stdout
    Messages._print('Generating report in', filename)

    # Generate the summary
    summary = _generate_summary(net_id)
    # Generate the populations
    populations = _generate_populations(net_id)
    # Generate the projections
    projections = _generate_projections(net_id, gather_subprojections)
    # Generate the neuron models
    neuron_models = _generate_neuron_models(net_id)
    # Generate the synapse models
    synapse_models = _generate_synapse_models(net_id)
    # Generate the constants
    constants = _generate_constants(net_id)
    # Generate the functions
    functions = _generate_functions(net_id)
    # Generate the population parameters
    pop_parameters = _generate_population_parameters(net_id)
    # Generate the population parameters
    proj_parameters = _generate_projection_parameters(net_id, gather_subprojections)
    # Generate the measurmements (in our case the BOLD recording)
    measurements = _generate_measurements(net_id)

    # Possibly create the directory if it does not exist
    path_name = os.path.dirname(filename)
    if not path_name in ["", "."]:
        if not os.path.exists(path_name):
            os.makedirs(path_name)

    with open(filename, 'w') as wfile:
        if standalone:
            wfile.write(header)
            wfile.write(preamble)
        wfile.write(summary)
        wfile.write(populations)
        wfile.write(projections)
        wfile.write(neuron_models)
        wfile.write(synapse_models)
        wfile.write(parameters_template)
        wfile.write(constants)
        wfile.write(functions)
        wfile.write(pop_parameters)
        wfile.write(proj_parameters)
        wfile.write(input_template)
        wfile.write(measurements)
        if standalone:
            wfile.write(footer)

##################################
### Process major fields
##################################

def _generate_summary(net_id):
    "part A"

    population_names = str(NetworkManager().number_populations(net_id=net_id)) + ': '
    connectivity = ""
    neuron_models = ""
    synapse_models = ""

    # List the names of all populations
    for pop in NetworkManager().get_populations(net_id=net_id):
        # population name
        population_names += LatexParser.pop_name(pop.name) + ", "
    population_names = population_names[:-2] # suppress the last ,

    # List all BOLD recordings
    measurements = ""

    # List all neuron types
    neuron_model_names = []
    for neur in GlobalObjectManager().get_neuron_types():
        # bold models sorted in measurements
        if isinstance(neur, BoldModel):
            if neur._model_instantiated:
                measurements += neur.name + ', '
        # neuron model
        else:
            neuron_model_names.append(neur.name)
    for neur in list(set(neuron_model_names)):
        neuron_models += neur + ', '
    # suppress the last ,
    measurements = measurements[:-2]
    neuron_models = neuron_models[:-2]

    list_connectivity = []
    list_synapse_models = []
    for proj in NetworkManager().get_projections(net_id=net_id):
        list_connectivity.append(proj.connector_name)
        if not proj.synapse_type.name in list(Synapse._default_names.values()) + ['-']:
            list_synapse_models.append(proj.synapse_type.name)
    for con in list(set(list_connectivity)):
        connectivity += con + ', '
    for syn in list(set(list_synapse_models)):
        synapse_models += syn + ', '
    connectivity = connectivity[:-2]
    synapse_models = synapse_models[:-2] # suppress the last ,

    # Write the summary
    txt = summary_template  % {
        'population_names' : population_names,
        'connectivity' : connectivity,
        'neuron_models' : neuron_models,
        'synapse_models' : synapse_models,
        'measurements': measurements
    }
    return txt

def _generate_populations(net_id):
    def format_size(pop):
        size = str(pop.size)
        if pop.dimension >1:
            size += ' ('
            for d in range(pop.dimension):
                size += str(pop.geometry[d]) + '*'
            size = size.rsplit('*', 1)[0] + ')'
        return size

    txt = ""
    pop_tpl = r"""
    %(pop_name)s             & %(neuron_type)s        & $N_{\text{%(pop_name)s}}$ = %(size)s  \\ \hline
"""
    for pop in NetworkManager().get_populations(net_id=net_id):
        # Find a name for the neuron
        if pop.neuron_type.name in Neuron._default_names.values(): # name not set
            neuron_name = "Neuron " + str(pop.neuron_type._rk_neurons_type)
        else:
            neuron_name = pop.neuron_type.name

        txt += pop_tpl % {
            'pop_name': LatexParser.pop_name(pop.name), 
            'neuron_type': neuron_name, 
            'size': format_size(pop)}

    return populations_template % {'populations_description': txt}

def _generate_constants(net_id):
    cst_tpl = r"""
    & $%(param)s$        & %(value)s  \\\\ \\hline
"""
    parameters = ""
    if GlobalObjectManager().number_constants() == 0:
        return ""

    for constant in GlobalObjectManager().get_constants():
        parameters += cst_tpl % {'param': LatexParser._latexify_name(constant.name, []), 'value': constant.value}

    txt = constants_template % {'parameters': parameters}

    return txt


def _generate_functions(net_id):

    functions = ""
    if GlobalObjectManager().number_functions() == 0:
        return functions

    for _, func in GlobalObjectManager().get_functions():
        functions += LatexParser._process_functions(func) + "\n"

    return functions_template % {
        'parameters': functions, 
        'firstfunction': r"\hdr{1}{G}{Functions}\\ \hline"}

def _generate_population_parameters(net_id):
    txt = ""
    pop_tpl = r"""
    %(name)s             & $%(param)s$        & %(value)s  \\ \hline
"""
    for rk, pop in enumerate(NetworkManager().get_populations(net_id=net_id)):
        parameters = ""
        for idx, param in enumerate(pop.parameters):
            val = pop.init[param]
            if isinstance(val, (list, np.ndarray)):
                val = "$[" + str(np.array(val).min()) + ", " + str(np.array(val).max()) + "]$"
            parameters += pop_tpl % {'name': LatexParser.pop_name(pop.name) if idx==0 else "", 'param': LatexParser._latexify_name(param, []), 'value': val}

        if parameters != "":
            txt += popparameters_template % {'parameters': parameters, 'firstpopulation': r"\hdr{3}{H}{Population parameters}\\ \hline" if rk==0 else ""}

    return txt

def _generate_projections(net_id, gather_subprojections):
    txt = r""
    proj_tpl = r"""
    %(pre)s & %(post)s & %(target)s & %(synapse)s &
    %(description)s \\ \hline
"""
    if gather_subprojections:
        projections = []
        for proj in NetworkManager().get_projections(net_id=net_id):
            for existing_proj in projections:
                if proj.pre.name == existing_proj.pre.name and proj.post.name == existing_proj.post.name and proj.target == existing_proj.target : # TODO
                    break
            else:
                projections.append(proj)
    else:
        projections = NetworkManager().get_projections(net_id=net_id)

    for proj in projections:
        # Find a name for the synapse
        if proj.synapse_type.name in Synapse._default_names.values(): # name not set
            synapse_name = "Synapse " + str(proj.synapse_type._rk_synapses_type)
        else:
            synapse_name = proj.synapse_type.name

        txt += proj_tpl % { 'pre': LatexParser.pop_name(proj.pre.name), 
                            'post': LatexParser.pop_name(proj.post.name), 
                            'target': LatexParser._format_list(proj.target, ' / '),
                            'synapse': synapse_name,
                            'description': proj.connector_description}

    return connectivity_template % {'projections_description': txt}

def _generate_projection_parameters(net_id, gather_subprojections):
    txt = ""
    proj_tpl = r"""
    %(name)s & $%(param)s$        & %(value)s  \\ \hline
"""
    if gather_subprojections:
        projections = []
        for proj in NetworkManager().get_projections(net_id=net_id):
            for existing_proj in projections:
                if proj.pre.name == existing_proj.pre.name and proj.post.name == existing_proj.post.name and proj.target == existing_proj.target : # TODO
                    break
            else:
                projections.append(proj)
    else:
        projections = NetworkManager().get_projections(net_id=net_id)

    first = True
    for rk, proj in enumerate(projections):
        parameters = ""
        for idx, param in enumerate(proj.parameters):
            if param == 'w':
                continue
            if idx == 0:
                proj_name = "%(pre)s  $\\rightarrow$ %(post)s with target %(target)s" % {
                    'pre': LatexParser.pop_name(proj.pre.name), 
                    'post': LatexParser.pop_name(proj.post.name), 
                    'target': LatexParser._format_list(proj.target, ' / ')}
            else:
                proj_name = ""
            val = proj.init[param]
            
            if isinstance(val, (list, np.ndarray)):
                val = "$[" + str(np.min(val)) + ", " + str(np.max(val)) + "]$"
            parameters += proj_tpl % {'name': proj_name, 'param': LatexParser._latexify_name(param, []), 'value': val}

        if parameters != "":
            txt += projparameters_template % {
                'parameters': parameters, 
                'firstprojection': r"\hdr{3}{H}{Projection parameters}\\ \hline" if first else ""}
            first = False

    return txt

def _generate_neuron_models(net_id):
    neurons = ""

    firstneuron = r"\hdr{2}{D}{Neuron Models}\\ \hline"

    neuron_tpl = r"""
\noindent
\begin{tabularx}{\linewidth}{|p{0.15\linewidth}|X|}\hline
%(firstneuron)s
\textbf{Name} & %(name)s \\ \hline
\textbf{Type} & %(description)s\\ \hline
\textbf{%(equation_type)s} &
%(variables)s
%(spike)s
%(functions)s
\end{tabularx}
\vspace{2ex}
"""

    first_neuron = True
    for neuron in GlobalObjectManager().get_neuron_types():

        # skip bold models
        if isinstance(neuron, BoldModel):
            continue

        # Name
        if neuron.name in Neuron._default_names.values(): # name not set
            neuron_name = "Neuron " + str(neuron._rk_neurons_type)
        else:
            neuron_name = neuron.name

        # Generate the code for the equations
        variables, spike_condition, spike_reset = LatexParser._process_neuron_equations(neuron)

        eqs = ""
        for var in variables:
            eqs += r"""
\begin{dmath*}
%(eq)s
\end{dmath*}
""" % {'eq': var['latex']}

        variables_eqs = r"""
%(eqs)s
\\ \hline
""" % {'eqs': eqs}

        # Spiking neurons have an extra field for the spike condition
        spike_extra = ""
        if neuron.type == 'spike':
            spike_code = "If $" + spike_condition + r"$ or $t \leq t^* + t_{\text{refractory}}$:"

            # Reset
            spike_code += r"""
            \begin{enumerate}
                \item Emit a spike at time $t^*$"""

            for var in spike_reset:
                spike_code += r"""
            \item $""" + var + "$"

            spike_code += r"""
        \end{enumerate}"""


            spike_extra = r"""
\textbf{Spiking} &
%(spike)s
\\ \hline
""" % {'spike': spike_code}

        # Possible function
        functions = ""
        if not neuron.functions == None:
            functions = r"""
\textbf{Functions} &
%(functions)s
\\ \hline
""" % {'functions': LatexParser._process_functions(neuron.functions)}

        # Build the dictionary
        desc = {
            'name': neuron_name,
            'description': neuron.short_description,
            'firstneuron': firstneuron if first_neuron else "",
            'variables': variables_eqs,
            'spike': spike_extra,
            'functions': functions,
            'equation_type': "Subthreshold dynamics" if neuron.type == 'spike' else 'Equations'
        }

        # Generate the code depending on the neuron position
        neurons += neuron_tpl % desc

        # first_neuron = False, to prevent multiple header
        first_neuron = False

    return neurons

def _generate_synapse_models(net_id):
    firstsynapse = ""
    synapses = ""

    firstsynapse = r"\hdr{2}{E}{Synapse Models}\\ \hline"

    synapse_tpl = r"""
\noindent
\begin{tabularx}{\linewidth}{|p{0.15\linewidth}|X|}\hline
%(firstsynapse)s
\textbf{Name} & %(name)s \\ \hline
\textbf{Type} & %(description)s\\ \hline
%(psp)s
%(variables)s
%(preevent)s
%(postevent)s
%(functions)s
\end{tabularx}
\vspace{2ex}
"""
    for idx, synapse in enumerate(GlobalObjectManager().get_synapse_types()):
        # Do not document default synapses
        if synapse.name == "-":
            continue

        # Find a name for the synapse
        if synapse.name in Synapse._default_names.values(): # name not set
            synapse_name = "Synapse " + str(synapse._rk_synapses_type)
        else:
            synapse_name = synapse.name

        # Generate the code for the equations
        psp, variables, pre_desc, post_desc = LatexParser._process_synapse_equations(synapse)

        eqs = ""
        for var in variables:
            eqs += r"""
\begin{dmath*}
%(eq)s
\end{dmath*}
""" % {'eq': var['latex']}

        variables_eqs = r"""
%(eqs)s
\\ \hline
""" % {'eqs': eqs}

        # Synaptic variables
        variables = r"""
\textbf{Equations} & %(variables)s  
\\ \hline""" % {'variables':eqs} if eqs != "" else ""

        # PSP
        if psp != "":
            psp_code = r"""
\textbf{PSP} & \begin{dmath*}
%(psp)s
\end{dmath*}
\\ \hline""" % {'psp': psp}
        else:
            psp_code = ""

        # Spiking neurons have extra fields for the event-driven
        if synapse.type == 'spike':
            if len(pre_desc) > 0:
                txt_pre = ""
                for l in pre_desc:
                    txt_pre += r"""
\begin{dmath*}
%(eq)s
\end{dmath*}
""" % {'eq': l}
                preevent = r"""
\textbf{Pre-synaptic event} &
%(preevent)s
\\ \hline
""" % {'preevent': txt_pre}
            else:
                preevent = ""

            if len(post_desc) > 0:
                txt_post = ""
                for l in post_desc:
                    txt_post += r"""
\begin{dmath*}
%(eq)s
\end{dmath*}
""" % {'eq': l}
                postevent = r"""
\textbf{Post-synaptic event} &
%(postevent)s
\\ \hline
""" % {'postevent': txt_post}
            else:
                postevent = ""
        else:
            preevent = ""
            postevent = ""

        # Possible functions
        functions = ""
        if not synapse.functions == None:
            functions = r"""
\textbf{Functions} &
%(functions)s
\\ \hline
""" % {'functions': LatexParser._process_functions(synapse.functions)}

        # Build the dictionary
        desc = {
            'name': synapse_name,
            'description': synapse.short_description,
            'firstsynapse': firstsynapse if idx == 0 else "",
            'variables': variables,
            'psp': psp_code,
            'preevent': preevent,
            'postevent': postevent,
            'functions': functions
        }

        # Generate the code
        synapses += synapse_tpl % desc

    return synapses

def _generate_measurements(net_id):
    measurements = ""

    header = r"\hdr{2}{D}{Bold Measurement Models}\\ \hline"

    measurements_template = r"""
\noindent
\begin{tabularx}{\linewidth}{|p{0.15\linewidth}|X|}\hline
%(header)s
\textbf{Name} & %(name)s \\ \hline
\textbf{Type} & %(description)s\\ \hline
\textbf{%(equation_type)s} &
%(variables)s
\end{tabularx}
\vspace{2ex}
"""

    first_define = True
    for neuron in GlobalObjectManager().get_neuron_types():

        # skip non-bold models
        if not isinstance(neuron, BoldModel):
            continue

        # the model is not used
        if not neuron._model_instantiated:
            continue

        # Name
        if neuron.name in Neuron._default_names.values(): # name not set
            neuron_name = "Neuron " + str(neuron._rk_neurons_type)
        else:
            neuron_name = neuron.name

        # Generate the code for the equations
        variables, _, _ = LatexParser._process_neuron_equations(neuron)

        eqs = ""
        for var in variables:
            eqs += r"""
\begin{dmath*}
%(eq)s
\end{dmath*}
""" % {'eq': var['latex']}

        variables_eqs = r"""
%(eqs)s
\\ \hline
""" % {'eqs': eqs}

        # Build the dictionary
        desc = {
            'name': neuron_name,
            'description': neuron.short_description,
            'header': header if first_define else "",
            'variables': variables_eqs,
            'equation_type': "Equations"
        }

        # Generate the code depending on the neuron position
        measurements += measurements_template % desc

        # first_define disable to prevent multiple table header
        first_define = False

    # no models were processed
    if len(measurements) == 0:
        measurements = r"""
\noindent
\begin{tabularx}{\linewidth}{|l|X|}\hline
\hdr{2}{H}{Measurement}\\ \hline
\textbf{Type} & \textbf{Description} \\ \hline
---
\\ \hline
\end{tabularx}
"""


    return measurements
